from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from dissect.target.plugins.apps.other.env import EnvironmentFilePlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_envfile(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    target_unix.add_plugin(EnvironmentFilePlugin)
    fs_unix.map_file("/root/foo.env", absolute_path("_data/plugins/apps/other/env/test.env"))
    fs_unix.map_file("/root/foo/bar/test.env", absolute_path("_data/plugins/apps/other/env/test.env"))

    records = list(target_unix.envfile(env_path="/root"))

    assert len(records) == 4 * 2

    assert isinstance(records[0].ts_mtime, datetime)
    assert records[0].key == "DEBUG"
    assert records[0].value == "True"
    assert records[0].comment is None
    assert records[0].path == "/root/foo.env"

    assert isinstance(records[1].ts_mtime, datetime)
    assert records[1].key == "OPTIONAL"
    assert records[1].value == ""
    assert records[1].comment == "comment"
    assert records[1].path == "/root/foo.env"

    assert isinstance(records[2].ts_mtime, datetime)
    assert records[2].key == "FOO"
    assert records[2].value == "bar"
    assert records[2].comment == "this is a comment"
    assert records[2].path == "/root/foo.env"

    assert isinstance(records[3].ts_mtime, datetime)
    assert records[3].key == "HELLO"
    assert records[3].value == "world"
    assert records[3].comment is None
    assert records[3].path == "/root/foo.env"



from __future__ import annotations

import datetime
import struct
from typing import TYPE_CHECKING

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey

NetworkHistoryRecord = TargetRecordDescriptor(
    "windows/registry/nethist",
    [
        ("datetime", "created"),
        ("datetime", "last_connected"),
        ("string", "profile_guid"),
        ("string", "profile_name"),
        ("string", "description"),
        ("string", "dns_suffix"),
        ("string", "first_network"),
        ("string", "default_gateway_mac"),
        ("string", "signature"),
    ],
)


class NethistPlugin(Plugin):
    """Windows network history plugin."""

    KEY = "HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Networklist\\Signatures"
    PROFILE_KEY = "HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Networklist\\Profiles"

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEY))) > 0:
            raise UnsupportedPluginError("No Networklist registry keys found")

    @export(record=NetworkHistoryRecord)
    def network_history(self) -> Iterator[NetworkHistoryRecord]:
        """Return attached network history.

        The HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Networklist\\Signatures and
        HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Networklist\\Profiles registry keys contain information
        about the networks to which the system has been connected, both wireless and wired.

        References:
            - https://web.archive.org/web/20221127181357/https://www.weaklink.org/2016/11/windows-network-profile-registry-keys/
        """
        target_tz = self.target.datetime.tzinfo

        for key in self.target.registry.keys(self.KEY):
            for kind in key.subkeys():
                for sig in kind.subkeys():
                    guid = sig.value("ProfileGuid").value
                    profile = self.find_profile(guid)

                    created = parse_ts(profile.value("DateCreated").value, tzinfo=target_tz)
                    last_connected = parse_ts(profile.value("DateLastConnected").value, tzinfo=target_tz)

                    yield NetworkHistoryRecord(
                        created=created,
                        last_connected=last_connected,
                        profile_guid=guid,
                        profile_name=profile.value("ProfileName").value,
                        description=sig.value("Description").value,
                        dns_suffix=sig.value("DnsSuffix").value,
                        first_network=sig.value("FirstNetwork").value,
                        default_gateway_mac=sig.value("DefaultGatewayMac").value.hex(),
                        signature=sig.name,
                        _target=self.target,
                    )

    def find_profile(self, guid: str) -> RegistryKey | None:
        for key in self.target.registry.keys(self.PROFILE_KEY):
            try:
                return key.subkey(guid)  # Just return the first one...
            except RegistryError:  # noqa: PERF203
                pass
        return None


def parse_ts(val: bytes, tzinfo: datetime.tzinfo = datetime.timezone.utc) -> datetime.datetime:
    items = list(struct.unpack("<8H", val))
    # If we remove the weekday (at position 2), this is a valid datetime tuple
    items.pop(2)
    return datetime.datetime(*items, tzinfo=tzinfo)



from __future__ import annotations

import logging
import sys
from typing import Any

TRACE_LEVEL = 5
logging.addLevelName(TRACE_LEVEL, "TRACE")

# For some reason, 3.10 and earlier need stacklevel=3
_STACK_LEVEL = 2 if sys.version_info >= (3, 11) else 3


class TraceLogger(logging.Logger):
    def trace(self, msg: Any, *args: Any, **kwargs: Any) -> None:
        if self.isEnabledFor(TRACE_LEVEL):
            # Set stacklevel so that logging shows the correct caller info
            kwargs.setdefault("stacklevel", _STACK_LEVEL)
            self._log(TRACE_LEVEL, msg, args, **kwargs)


logging.setLoggerClass(TraceLogger)


def get_logger(name: str | None = None) -> TraceLogger:
    """Get a logger with ``TRACE`` support."""
    return logging.getLogger(name)


class TargetLogAdapter(logging.LoggerAdapter):
    def process(self, msg: str, kwargs: dict[str, Any]) -> tuple[str, dict[str, Any]]:
        return f"{self.extra['target']}: {msg}", kwargs

    def trace(self, msg: Any, *args, **kwargs) -> None:
        """Delegate a trace call to the underlying logger."""
        # Set stacklevel so that logging shows the correct caller info
        self.log(TRACE_LEVEL, msg, *args, stacklevel=_STACK_LEVEL, **kwargs)



from __future__ import annotations

import re
import urllib.parse
from typing import TYPE_CHECKING, Any, Union, get_args

from dissect.database.ese import ESE
from dissect.database.sqlite3 import SQLite3
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import FilesystemError, UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.apps.browser.browser import BrowserHistoryRecord
from dissect.target.plugins.apps.browser.edge import EdgePlugin
from dissect.target.plugins.apps.browser.iexplore import InternetExplorerPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.database.ese.record import Record as EseRecord
    from dissect.database.ese.table import Table as EseTable

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target

SearchIndexRecord = TargetRecordDescriptor(
    "windows/search/index/entry",
    [
        ("datetime", "ts"),
        ("datetime", "ts_mtime"),
        ("datetime", "ts_btime"),
        ("datetime", "ts_atime"),
        ("path", "path"),
        ("string", "type"),
        ("filesize", "size"),
        ("string", "data"),
        ("path", "source"),
    ],
)

SearchIndexActivityRecord = TargetRecordDescriptor(
    "windows/search/index/activity",
    [
        ("datetime", "ts_start"),
        ("datetime", "ts_end"),
        ("varint", "duration"),
        ("string", "application_name"),
        ("string", "application_id"),
        ("string", "activity_id"),
        ("path", "source"),
    ],
)

RE_URL = re.compile(r"(?P<browser>.+)\:\/\/\{(?P<sid>.+)\}\/(?P<url>.+)$")

BROWSER_RECORD_MAP = {
    "iehistory": InternetExplorerPlugin.BrowserHistoryRecord,
    "winrt": EdgePlugin.BrowserHistoryRecord,
}

SearchIndexRecords = Union[SearchIndexRecord, SearchIndexActivityRecord, BrowserHistoryRecord]  # noqa: UP007


class SearchIndexPlugin(Plugin):
    """Windows Search Index plugin."""

    SYSTEM_PATHS = (
        # Windows 11 22H2 (SQLite3)
        "sysvol/ProgramData/Microsoft/Search/Data/Applications/Windows/Windows.db",
        # Windows Vista and Windows 10 (ESE)
        "sysvol/ProgramData/Microsoft/Search/Data/Applications/Windows/Windows.edb",
        # Windows XP (ESE)
        "sysvol/Documents and Settings/All Users/Application Data/Microsoft/Search/Data/Applications/Windows/Windows.edb",  # noqa: E501
    )

    USER_PATHS = (
        # Windows 10 Server Roaming (ESE / SQLite3)
        "AppData/Roaming/Microsoft/Search/Data/Applications/S-1-*/*.*db",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.databases = set(self.find_databases())

    def find_databases(self) -> Iterator[tuple[Path, UserDetails | None]]:
        seen = set()

        for system_path in self.SYSTEM_PATHS:
            if (path := self.target.fs.path(system_path)).is_file():
                try:
                    if any(seen_file.samefile(path) for seen_file in seen):
                        continue
                except FilesystemError:
                    pass

                seen.add(path)
                yield path.resolve(), None

        for user_details in self.target.user_details.all_with_home():
            for user_path in self.USER_PATHS:
                for path in user_details.home_path.glob(user_path):
                    if not path.is_file():
                        continue
                    try:
                        if any(seen_file.samefile(path) for seen_file in seen):
                            continue
                    except FilesystemError:
                        pass

                    seen.add(path)
                    yield path.resolve(), user_details

    def check_compatible(self) -> None:
        if not self.databases:
            raise UnsupportedPluginError("No Windows Search Index database files found on target")

    @export(record=get_args(SearchIndexRecords))
    def search(self) -> Iterator[SearchIndexRecords]:
        """Yield Windows Search Index records.

        Parses ``Windows.edb`` ESE and ``Windows.db`` SQLite3 databases. Currently does not parse
        ``GatherLogs/SystemIndex/SystemIndex.*.(Crwl|gthr)`` files or ``Windows-gather.db`` and ``Windows-usn.db`` files.

        Windows Search is a standard component of Windows 7 and Windows Vista, and is enabled by default. The standard (non-Windows Server)
        configuration of Windows Search indexes the following paths: ``C:\\Users\\*`` and ``C:\\ProgramData\\Microsoft\\Windows\\Start Menu\\Programs\\*``,
        with some exceptions for certain file extensions (see the linked references for more information).

        The difference between the fields ``System_Date*`` and ``System_Document_Date*`` should be researched further.
        It is unclear what the field ``InvertedOnlyMD5`` is a checksum of (record or file content?). It might be possible
        to correlate the field ``System_FileOwner`` with a ``UserRecordDescriptor``. The field ``System_FileAttributes`` should be
        investigated further.

        No test data available for indexed Outlook emails, this plugin might not be able to handle indexed email messages.

        References:
            - https://learn.microsoft.com/en-us/windows/win32/search/-search-3x-wds-overview
            - https://github.com/libyal/esedb-kb/blob/main/documentation/Windows%20Search.asciidoc
            - https://www.aon.com/en/insights/cyber-labs/windows-search-index-the-forensic-artifact-youve-been-searching-for
            - https://github.com/strozfriedberg/sidr
            - https://devblogs.microsoft.com/windows-search-platform/configuration-and-settings/
            - https://learn.microsoft.com/en-us/windows/win32/search/-search-3x-wds-included-in-index
        """  # noqa: E501

        for db_path, user_details in self.databases:
            if db_path.suffix == ".edb":
                yield from self.parse_esedb(db_path, user_details)

            elif db_path.suffix == ".db":
                yield from self.parse_sqlite(db_path, user_details)

            else:
                self.target.log.warning("Unknown Windows Search Index database file %r", db_path)

    def parse_esedb(self, path: Path, user_details: UserDetails | None) -> Iterator[SearchIndexRecords]:
        """Parse the ESE ``SystemIndex_PropertyStore`` table."""

        with path.open("rb") as fh:
            db = ESE(fh)
            table = db.table("SystemIndex_PropertyStore")

            for record in table.records():
                yield from self.build_record(TableRecord(table, record), user_details, path)

    def parse_sqlite(self, path: Path, user_details: UserDetails | None) -> Iterator[SearchIndexRecords]:
        """Parse the SQLite3 ``SystemIndex_1_PropertyStore`` table."""

        with SQLite3(path) as db:
            # ``ColumnId`` is translated using the ``SystemIndex_1_PropertyStore_Metadata`` table.
            columns = {
                row.get("Id"): row.get("UniqueKey", "").split("-", maxsplit=1)[-1]
                for row in db.table("SystemIndex_1_PropertyStore_Metadata").rows()
            }

            if not (table := db.table("SystemIndex_1_PropertyStore")):
                self.target.log.warning("Database %s does not have a table called 'SystemIndex_1_PropertyStore'", path)
                return

            current_work_id = None
            values = {}

            for row in table.rows():
                work_id = row.get("WorkId")
                if current_work_id is None:
                    current_work_id = work_id
                if work_id != current_work_id:
                    yield from self.build_record(values, user_details, path)
                    current_work_id = work_id
                    values = {}

                if value := row.get("Value"):
                    column_name = columns[row.get("ColumnId")]
                    values[column_name] = value

            yield from self.build_record(values, user_details, path)

    def build_record(
        self, values: dict[str, Any] | TableRecord, user_details: UserDetails | None, db_path: Path
    ) -> Iterator[SearchIndexRecords]:
        """Build a ``SearchIndexRecord``, ``SearchIndexActivityRecord`` or ``HistoryRecord``."""

        if values.get("System_ItemType") == "ActivityHistoryItem":
            yield SearchIndexActivityRecord(
                ts_start=wintimestamp(int.from_bytes(values.get("System_ActivityHistory_StartTime", b""), "little")),
                ts_end=wintimestamp(int.from_bytes(values.get("System_ActivityHistory_EndTime", b""), "little")),
                duration=int.from_bytes(values.get("System_ActivityHistory_ActiveDuration", b""), "little"),
                application_name=values.get("System_Activity_AppDisplayName"),
                application_id=values.get("System_ActivityHistory_AppId"),
                activity_id=values.get("System_ActivityHistory_AppActivityId"),
                source=db_path,
                _target=self.target,
            )

        elif values.get("System_Search_Store") in ("iehistory", "winrt"):
            system_itemurl = values.get("System_ItemUrl")

            if not system_itemurl or not (match := RE_URL.match(system_itemurl)):
                self.target.log.warning(
                    "Unable to parse System_ItemUrl: %r (%r) in %s", system_itemurl, values, db_path
                )
                return

            browser, sid, url = match.groupdict().values()

            if not (CurrentBrowserHistoryRecord := BROWSER_RECORD_MAP.get(browser)):
                self.target.log.warning(
                    "Unable to determine browser history type for %r (%r) in %s", browser, system_itemurl, db_path
                )
                return

            user = None
            if sid and (sid_user_details := self.target.user_details.find(sid)):
                user = sid_user_details.user

            if not user and user_details:
                user = user_details.user

            url = values.get("System_Link_TargetUrl") or url
            host = None

            if url:
                try:
                    host = urllib.parse.urlparse(url).hostname
                except Exception:
                    pass

            yield CurrentBrowserHistoryRecord(
                ts=wintimestamp(int.from_bytes(values.get("System_Link_DateVisited", b""), "little")),
                browser=browser,
                url=url,
                title=values.get("System_Title"),
                host=host,
                source=db_path,
                _user=user,
                _target=self.target,
            )

        # System_Search_Store = "file"
        else:
            yield SearchIndexRecord(
                ts=wintimestamp(int.from_bytes(values.get("System_Search_GatherTime", b""), "little")),
                ts_mtime=wintimestamp(int.from_bytes(values.get("System_DateModified", b""), "little")),
                ts_btime=wintimestamp(int.from_bytes(values.get("System_DateCreated", b""), "little")),
                ts_atime=wintimestamp(int.from_bytes(values.get("System_DateAccessed", b""), "little")),
                path=values.get("System_ItemPathDisplay"),
                type=values.get("System_MIMEType")
                or values.get("System_ContentType")
                or values.get("System_ItemTypeText"),
                size=int.from_bytes(b_size, "little") if (b_size := values.get("System_Size")) else None,
                data=values.get("System_Search_AutoSummary"),
                source=db_path,
                _target=self.target,
            )


class TableRecord:
    def __init__(self, table: EseTable, record: EseRecord):
        self.table = table
        self.record = record

        # Translates e.g. ``System_DateModified`` to ``15F-System_DateModified`` as these column name prefixes might
        # be dynamic based on the system version.
        self.columns = {col.split("-", maxsplit=1)[-1]: col for col in table.column_names}

    def get(self, key: str, default: Any | None = None) -> Any:
        return self.record.get(self.columns.get(key, default))



from __future__ import annotations

from dissect.target.plugins.os.unix.bsd.citrix.locale import CitrixLocalePlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_locale_nscollector() -> None:
    """Test that the Citrix locale plugin works as expected on NetScaler Collector packages."""
    path = absolute_path("_data/loaders/nscollector/collector_P_10.164.0.3_22Oct2025_11_31.tar.gz")

    t = Target.open(path)
    t.add_plugin(CitrixLocalePlugin)

    assert t.timezone == "UTC"
    assert t.language == []



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.filesystems.tar import TarFilesystem
from dissect.target.helpers.logging import get_logger
from dissect.target.loader import Loader
from dissect.target.loaders.dir import find_and_map_dirs, find_dirs, map_dirs
from dissect.target.loaders.tar import TarSubLoader
from dissect.target.loaders.zip import ZipSubLoader

if TYPE_CHECKING:
    import tarfile as tf
    import zipfile as zf
    from pathlib import Path

    from dissect.target.target import Target


log = get_logger(__name__)

FILESYSTEMS_ROOT = "[root]"
UAC_CHECK_FILE = "uac.log"


class UacLoader(Loader):
    """Loader for extracted UAC collections.

    References:
        - https://github.com/tclahr/uac
        - https://tclahr.github.io/uac-docs/
    """

    @staticmethod
    def detect(path: Path) -> bool:
        return path.joinpath(FILESYSTEMS_ROOT).exists() and path.joinpath(UAC_CHECK_FILE).exists()

    def map(self, target: Target) -> None:
        os_type, dirs = find_dirs(self.absolute_path.joinpath(FILESYSTEMS_ROOT))
        map_dirs(target, dirs, os_type)


class UacTarSubloader(TarSubLoader):
    """Loader for tar-based UAC collections.

    References:
        - https://github.com/tclahr/uac
        - https://tclahr.github.io/uac-docs/
    """

    FS_ROOT_TUPLE = (f"/{FILESYSTEMS_ROOT}/", f"{FILESYSTEMS_ROOT}/", f"./{FILESYSTEMS_ROOT}/")

    @staticmethod
    def detect(path: Path, tarfile: tf.TarFile) -> bool:
        return any(member.name.startswith(UacTarSubloader.FS_ROOT_TUPLE) for member in tarfile.getmembers())

    def map(self, target: Target) -> None:
        vol = TarFilesystem(tarfile=self.tar, base=FILESYSTEMS_ROOT, fh=self.tar.fileobj)
        target.filesystems.add(vol)


class UacZipSubLoader(ZipSubLoader):
    """Loader for zip-based UAC collections.

    References:
        - https://github.com/tclahr/uac
        - https://tclahr.github.io/uac-docs/
    """

    @staticmethod
    def detect(path: Path, zipfile: zf.Path) -> bool:
        return zipfile.joinpath(FILESYSTEMS_ROOT).exists() and zipfile.joinpath(UAC_CHECK_FILE).exists()

    def map(self, target: Target) -> None:
        path = self.zip.joinpath(FILESYSTEMS_ROOT)
        find_and_map_dirs(target, path)



from __future__ import annotations

from typing import TYPE_CHECKING, Final

from dissect.target.exceptions import RegistryError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import (
    RegistryRecordDescriptorExtension,
    UserRecordDescriptorExtension,
)
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.regutil import RegistryKey
    from dissect.target.target import Target

CLSIDRecordDescriptor = create_extended_descriptor(
    [
        RegistryRecordDescriptorExtension,
        UserRecordDescriptorExtension,
    ]
)

CLSIDRecord = CLSIDRecordDescriptor(
    "windows/registry/clsid",
    [
        ("datetime", "ts"),
        ("string", "clsid"),
        ("string", "name"),
        ("string", "value"),
    ],
)


class CLSIDPlugin(Plugin):
    """Return all CLSID registry keys.

    A CLSID is a globally unique identifier that identifies a COM class object (program) situated in
    HKEY_CURRENT_USER\\Software\\Classes\\CLSID and HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\CLSID. Malware may make use
    of the CLSID system to launch themselves automatically or when certain conditions are triggered.

    References:
        - https://docs.microsoft.com/en-us/windows/win32/com/clsid-key-hklm
        - https://www.enigmasoftware.com/what-is-clsid-registry-key/
    """

    __namespace__ = "clsid"

    KEYS: Final[dict[str, str]] = {
        "user": "HKEY_CURRENT_USER\\Software\\Classes\\CLSID",
        "machine": "HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\CLSID",
    }

    def __init__(self, target: Target):
        super().__init__(target)

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(list(self.KEYS.values())))) > 0:
            raise UnsupportedPluginError("No CLSID key found")

    def create_records(self, keys: list[RegistryKey]) -> Iterator[CLSIDRecord]:
        """Iterate all CLSID keys from HKEY_CURRENT_USER\\Software\\Classes\\CLSID and
        HKEY_LOCAL_MACHINE\\SOFTWARE\\Classes\\CLSID.

        Yields CLSIDRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): Last modified timestamp of the registry key.
            clsid (string): The CLSID key name.
            path (uri): The CLSID path value.
        """
        names = [
            "InprocServer32",
            "InprocServer",
            "LocalServer",
            "LocalServer32",
        ]

        for reg in self.target.registry.keys(keys):
            user = self.target.registry.get_user(reg)

            for subkey in reg.subkeys():
                try:
                    name = subkey.value("(default)").value
                except RegistryError:
                    name = None

                for entry in subkey.subkeys():
                    if entry.name in names:
                        try:
                            subkey_value = entry.value("(default)")
                        except RegistryError:
                            continue

                        yield CLSIDRecord(
                            ts=entry.ts,
                            clsid=subkey.name,
                            name=name,
                            value=subkey_value.value,
                            _target=self.target,
                            _user=user,
                            _key=entry,
                        )

    @export(record=CLSIDRecord)
    def user(self) -> Iterator[CLSIDRecord]:
        """Return only the user CLSID registry keys."""
        yield from self.create_records(self.KEYS["user"])

    @export(record=CLSIDRecord)
    def machine(self) -> Iterator[CLSIDRecord]:
        """Return only the machine CLSID registry keys."""
        yield from self.create_records(self.KEYS["machine"])



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.child.vmware_workstation import (
    VmwareWorkstationChildTargetPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("target_fixture", "fs_fixture", "inventory_path"),
    [
        pytest.param(
            "target_win_users",
            "fs_win",
            "Users\\John\\AppData\\Roaming\\VMware\\inventory.vmls",
            id="windows",
        ),
        pytest.param(
            "target_unix_users",
            "fs_unix",
            "/home/user/.vmware/inventory.vmls",
            id="linux",
        ),
        pytest.param(
            "target_macos_users",
            "fs_macos",
            "/Users/dissect/Library/Application Support/VMware Fusion/vmInventory",
            id="macos",
        ),
    ],
)
def test_child_vmware_workstation(
    target_fixture: str, fs_fixture: str, inventory_path: str, request: pytest.FixtureRequest
) -> None:
    """Test if we detect VMware Workstation children from inventory files correctly on Windows and Unix targets."""

    target: Target = request.getfixturevalue(target_fixture)
    fs: VirtualFilesystem = request.getfixturevalue(fs_fixture)

    fs.map_file(inventory_path, absolute_path("_data/plugins/child/vmware_workstation/inventory.vmls"))

    target.add_plugin(VmwareWorkstationChildTargetPlugin)
    children = [child for _, child in target.list_children()]

    assert len(children) == 3
    assert children[0].type == "vmware_workstation"
    assert children[0].name == "First Virtual Machine"
    assert children[0].path == "/path/to/first/vm/vm.vmx"

    assert [c.path for c in children] == [
        "/path/to/first/vm/vm.vmx",
        "/path/to/second/vm/vm.vmx",
        "/path/to/third/vm/vm.vmx",
    ]



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import patch

import pytest

from dissect.target.plugins.apps.remoteaccess.splashtop import SplashtopPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


SPLASHTOP_LOG_PATH = "Program Files (x86)/Splashtop/Splashtop Remote/Server/log/SPLog.txt"


@pytest.fixture
def target_splashtop(target_win_users: Target, fs_win: VirtualFilesystem) -> Iterator[Target]:
    fs_win.map_file(
        SPLASHTOP_LOG_PATH,
        absolute_path("_data/plugins/apps/remoteaccess/splashtop/SPLog.txt"),
    )

    # The Splashtop plugin uses a year rollover helper which uses the modification time of a file to determine
    # the starting year
    # A new source checkout would result in different modification timestamps, so mock it to be in 2025
    entry = fs_win.get(SPLASHTOP_LOG_PATH)
    stat_result = entry.stat()
    stat_result.st_mtime = 1735732800

    with patch.object(entry, "stat") as mock_stat:
        mock_stat.return_value = stat_result
        target_win_users.add_plugin(SplashtopPlugin)
        yield target_win_users


def test_splashtop_plugin_log(target_splashtop: Target) -> None:
    records = list(target_splashtop.splashtop.logs())
    assert len(records) == 384

    assert records[-1].ts == datetime(2025, 7, 14, 15, 15, 38, 194000, tzinfo=timezone.utc)
    assert records[-1].message == "SM_03280[Network] [LAN-S][Server] client connected from 10.199.5.134 (2), 288"
    assert records[-1].source == f"sysvol/{SPLASHTOP_LOG_PATH}"


def test_splashtop_plugin_filetransfer(target_splashtop: Target) -> None:
    records = list(target_splashtop.splashtop.filetransfer())
    assert len(records) == 1

    assert records[0].ts == datetime(2025, 7, 14, 15, 17, 30, 766000, tzinfo=timezone.utc)
    assert (
        records[0].message
        == 'SM_03280[FTCnnel] OnUploadFileCPRequest 1, 1 =>{"fileID":"353841253","fileName":"NOTE.txt","fileSize":"34","remotesessionFTC":1,"request":"uploadFile"}'  # noqa: E501
    )
    assert records[0].source == f"sysvol/{SPLASHTOP_LOG_PATH}"
    assert records[0].filename == "NOTE.txt"



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

CronjobRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "unix/cronjob",
    [
        ("string", "minute"),
        ("string", "hour"),
        ("string", "day"),
        ("string", "month"),
        ("string", "weekday"),
        ("string", "command"),
        ("path", "source"),
    ],
)

EnvironmentVariableRecord = TargetRecordDescriptor(
    "unix/environmentvariable",
    [
        ("string", "key"),
        ("string", "value"),
        ("path", "source"),
    ],
)

RE_CRONJOB = re.compile(
    r"""
        ^
        (?P<minute>\S+)
        \s+
        (?P<hour>\S+)
        \s+
        (?P<day>\S+)
        \s+
        (?P<month>\S+)
        \s+
        (?P<weekday>\S+)
        \s+
        (?P<command>.+)
        $
    """,
    re.VERBOSE,
)
RE_ENVVAR = re.compile(r"^(?P<key>[a-zA-Z_]+[a-zA-Z[0-9_])=(?P<value>.*)")


class CronjobPlugin(Plugin):
    """Unix cronjob plugin."""

    CRONTAB_DIRS = (
        "/var/cron/tabs",
        "/var/spool/cron",
        "/var/spool/cron/crontabs",
        "/etc/cron.d",
        "/usr/local/etc/cron.d",  # FreeBSD
    )

    CRONTAB_FILES = (
        "/etc/crontab",
        "/etc/anacrontab",
    )

    def __init__(self, target: Target):
        super().__init__(target)
        self.crontabs = list(self.find_crontabs())

    def check_compatible(self) -> None:
        if not self.crontabs:
            raise UnsupportedPluginError("No crontab(s) found on target")

    def find_crontabs(self) -> Iterator[Path]:
        for crontab_dir in self.CRONTAB_DIRS:
            if not (dir := self.target.fs.path(crontab_dir)).exists():
                continue

            for file in dir.iterdir():
                if file.resolve().is_file():
                    yield file

        for crontab_file in self.CRONTAB_FILES:
            if (file := self.target.fs.path(crontab_file)).exists():
                yield file

    @export(record=[CronjobRecord, EnvironmentVariableRecord])
    def cronjobs(self) -> Iterator[CronjobRecord | EnvironmentVariableRecord]:
        """Yield cronjobs, and their configured environment variables on a Unix system

        A cronjob is a scheduled task/command on a Unix based system. Adversaries may use cronjobs to gain
        persistence on the system.

        References:
            - https://linux.die.net/man/8/cron
            - https://linux.die.net/man/1/crontab
            - https://linux.die.net/man/5/crontab
            - https://en.wikipedia.org/wiki/Cron
            - https://linux.die.net/man/8/anacron
            - https://manpages.ubuntu.com/manpages/oracular/en/man5/crontab.5.html
            - https://www.gnu.org/software/mcron/manual/mcron.html#Guile-Syntax
        """

        for file in self.crontabs:
            # Cronjobs in user crontab files do not have a user field specified.
            user = None
            if file.is_relative_to("/var/spool/cron/crontabs") or file.is_relative_to("/var/spool/cron/"):
                user = self.target.user_details.find(username=file.name)

            for line in file.open("rt"):
                line = line.strip()
                if line.startswith("#") or not line:
                    continue

                if match := RE_CRONJOB.search(line):
                    match = match.groupdict()

                    # We try to infer a possible user from the command. This can lead to false positives,
                    # due to differing implementations of cron across operating systems, which is why
                    # we choose not to change the 'command' from the cron line - unless the command
                    # starts with the found username plus a tab character. We try to weed out false
                    # positives by checking the inferred user with the target's users.
                    if not user:
                        try:
                            inferred_user, _ = re.split(r"\s", match["command"], maxsplit=1)
                            # If the inferred username exists on the target we assign that user to this cronjob.
                            if user := self.target.user_details.find(username=inferred_user.strip()):
                                pass

                            # If the inferred username is followed by a tab we remove the username from the command.
                            if match["command"].startswith(inferred_user + "\t"):
                                match["command"] = match["command"].replace(inferred_user + "\t", "", 1)
                        except ValueError:
                            pass

                    yield CronjobRecord(
                        **match,
                        source=file,
                        _user=user.user if user else None,
                        _target=self.target,
                    )

                # Some cron implementations allow for environment variables to be set inside crontab files.
                elif match := RE_ENVVAR.search(line):
                    match = match.groupdict()
                    yield EnvironmentVariableRecord(
                        **match,
                        source=file,
                        _target=self.target,
                    )

                else:
                    self.target.log.warning("Unable to match cronjob line in %s: '%s'", file, line)



from __future__ import annotations

from typing import Any

from dissect.target.exceptions import PluginError


class RecordDescriptorExtensionBase:
    _default_fields = ()
    _input_fields = ()
    _prepend_fields = False

    def _fill_default_fields(self, record_kwargs: dict[str, Any]) -> dict[str, Any]:
        raise NotImplementedError


class UserRecordDescriptorExtension(RecordDescriptorExtensionBase):
    _default_fields = (
        ("string", "username"),
        ("string", "user_id"),
        ("string", "user_group"),
        ("string", "user_home"),
    )

    _input_fields = ("_user",)

    def _fill_default_fields(self, record_kwargs: dict[str, Any]) -> dict[str, Any]:
        user = record_kwargs.get("_user")

        username = None
        user_id = None
        user_group = None
        user_home = None

        if user:
            username = user.name
            user_id = getattr(user, "sid", None)
            if user_id is None:
                user_id = getattr(user, "uid", None)
            user_group = getattr(user, "gid", None)
            user_home = user.home

        record_kwargs.update(
            {"username": username, "user_id": user_id, "user_group": user_group, "user_home": user_home}
        )
        return record_kwargs


class RegistryRecordDescriptorExtension(RecordDescriptorExtensionBase):
    _default_fields = (
        ("string", "regf_hive_path"),
        ("string", "regf_key_path"),
    )

    _input_fields = ("_key", "_hive")

    def _fill_default_fields(self, record_kwargs: dict[str, Any]) -> dict[str, Any]:
        key = record_kwargs.get("_key")
        hive = record_kwargs.get("_hive", key.hive if hasattr(key, "hive") else None)
        record_kwargs["regf_key_path"] = key.path if hasattr(key, "path") else None
        record_kwargs["regf_hive_path"] = hive.filepath if hasattr(hive, "filepath") else None
        return record_kwargs


class TargetRecordDescriptorExtension(RecordDescriptorExtensionBase):
    _default_fields = (
        ("string", "hostname"),
        ("string", "domain"),
    )

    _input_fields = ("_target",)

    _prepend_fields = True

    def _fill_default_fields(self, record_kwargs: dict[str, Any]) -> dict[str, Any]:
        hostname = None
        domain = None
        source = None

        target = record_kwargs.get("_target")
        if target:
            try:
                hostname = target.hostname
            except PluginError:
                pass

            try:
                domain = target.domain
            except PluginError:
                pass
            source = target.path

        record_kwargs["hostname"] = hostname
        record_kwargs["domain"] = domain
        # Reserved keywords are never part of the args of a Record's __init__.
        record_kwargs["_source"] = source
        return record_kwargs



from __future__ import annotations

import ipaddress
from functools import cached_property
from typing import TYPE_CHECKING

from dissect.util import ts

from dissect.target.exceptions import (
    LoaderError,
    RegistryError,
    RegistryKeyNotFoundError,
    RegistryValueNotFoundError,
)
from dissect.target.filesystems.cb import CbFilesystem
from dissect.target.helpers.fsutil import TargetPath
from dissect.target.helpers.regutil import (
    RegistryHive,
    RegistryKey,
    RegistryValue,
    ValueType,
)
from dissect.target.loader import Loader
from dissect.target.plugins.os.windows.registry import RegistryPlugin

if TYPE_CHECKING:
    from datetime import datetime
    from pathlib import Path
    from urllib.parse import ParseResult

    from cbc_sdk.live_response_api import LiveResponseSession

    from dissect.target.target import Target

try:
    from cbc_sdk.errors import CredentialError
    from cbc_sdk.platform import Device
    from cbc_sdk.rest_api import CBCloudAPI

    HAS_CARBON_BLACK = True
except ImportError:
    HAS_CARBON_BLACK = False


class CbLoader(Loader):
    """Use Carbon Black endpoints as targets using Live Response.

    Use as ``cb://<hostname or IP>[@<instance>]``.

    Refer to the Carbon Black documentation for setting up a ``credentials.cbc`` file.
    """

    def __init__(self, path: Path, parsed_path: ParseResult | None = None):
        super().__init__(path, parsed_path=parsed_path, resolve=False)

        if not HAS_CARBON_BLACK:
            raise ImportError(
                "Required dependency 'carbon-black-cloud-sdk-python' is missing, install with 'pip install dissect.target[cb]'"  # noqa: E501
            )

        self.host = self.parsed_path.path.strip("/")
        instance = self.parsed_path.hostname

        try:
            self.cbc_api = CBCloudAPI(profile=instance or None)
        except CredentialError:
            raise LoaderError("The Carbon Black Cloud API key was not found or has the wrong permissions set")

        self.sensor = self.get_device()
        if not self.sensor:
            raise LoaderError("The device was not found within the specified instance")

        self.session = self.sensor.lr_session()

    def get_device(self) -> Device | None:
        try:
            ipaddress.ip_address(self.host)
            host_is_ip = True
        except ValueError:
            host_is_ip = False

        for cbc_sensor in self.cbc_api.select(Device).all():
            if host_is_ip:
                if cbc_sensor.last_internal_ip_address == self.host:
                    return cbc_sensor
            else:
                if device_name := getattr(cbc_sensor, "name", None):
                    device_name = device_name.lower()

                    # Sometimes the domain name is included in the device name
                    # E.g. DOMAIN\\Hostname
                    if "\\" in device_name:
                        device_name = device_name.split("\\")[1]

                    if device_name == self.host.lower():
                        return cbc_sensor

        return None

    @staticmethod
    def detect(path: Path) -> bool:
        return False

    def map(self, target: Target) -> None:
        for drive in self.session.session_data["drives"]:
            cbfs = CbFilesystem(self.session, drive)

            target.filesystems.add(cbfs)
            target.fs.mount(drive.lower(), cbfs)

        target.add_plugin(CbRegistry(target, self.session), check_compatible=False)


class CbRegistry(RegistryPlugin):
    __register__ = False

    def __init__(self, target: Target, session: LiveResponseSession):
        self.session = session
        super().__init__(target)

    def check_compatible(self) -> bool:
        return False

    def _init_registry(self) -> None:
        for hive_name, root_key in self.MAPPINGS.items():
            try:
                hive = CbRegistryHive(self.session, root_key)
                self._add_hive(hive_name, hive, TargetPath(self.target.fs, "CBR"))
                self._map_hive(root_key, hive)
            except RegistryError:  # noqa: PERF203
                continue


class CbRegistryHive(RegistryHive):
    def __init__(self, session: LiveResponseSession, root_key: str):
        self.session = session
        self.root_key = root_key

    def key(self, key: str) -> CbRegistryKey:
        path = f"{self.root_key}\\{key}" if key else self.root_key
        return CbRegistryKey(self, path)


class CbRegistryKey(RegistryKey):
    def __init__(self, hive: CbRegistryHive, path: str):
        self.session = hive.session
        self._path: str = path
        self._name: str = path.rsplit("\\", 1)[-1]
        super().__init__(hive)

    @cached_property
    def data(self) -> dict:
        try:
            return self.session.list_registry_keys_and_values(self._path)
        except Exception:
            raise RegistryKeyNotFoundError(self.path)

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path

    @property
    def timestamp(self) -> datetime:
        return ts.from_unix(0)

    def subkey(self, subkey: str) -> CbRegistryKey:
        # To improve peformance, immediately return a "hollow" key object
        # Only listing all subkeys or reading a value will result in data being loaded
        # Technically this means we won't raise a RegistryKeyNotFoundError in the correct place
        return CbRegistryKey(self.hive, f"{self._path}\\{subkey}")

    def subkeys(self) -> list[CbRegistryKey]:
        return list(map(self.subkey, self.data["sub_keys"]))

    def _value(self, value: str) -> CbRegistryValue:
        reg_value = value.lower()
        for val in self.values():
            if val.name.lower() == reg_value:
                return val
        else:
            raise RegistryValueNotFoundError(value)

    def values(self) -> list[CbRegistryValue]:
        return [
            CbRegistryValue(self.hive, val["registry_name"], val["registry_data"], val["registry_type"])
            for val in self.data["values"]
        ]


class CbRegistryValue(RegistryValue):
    def __init__(self, hive: CbRegistryHive, name: str, data: str, type: str):
        super().__init__(hive)
        self._name = name
        self._type = type

        if self._type == "pbREG_BINARY":
            self._value = bytes.fromhex(data)
        elif self._type in ("pbREG_DWORD", "pbREG_QWORD", "pbREG_DWORD_BIG_ENDIAN"):
            self._value = int(data)
        elif self._type == "pbREG_MULTI_SZ":
            self._value = data.split(",")
        else:
            # pbREG_NONE, pbREG_SZ, pbREG_EXPAND_SZ
            self._value = data

    @property
    def name(self) -> str:
        return self._name

    @property
    def value(self) -> ValueType:
        return self._value

    @property
    def type(self) -> str:
        return self._type



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_COOKIE_FIELDS,
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_EXTENSION_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    GENERIC_PASSWORD_RECORD_FIELDS,
    BrowserPlugin,
)
from dissect.target.plugins.apps.browser.chromium import (
    CHROMIUM_DOWNLOAD_RECORD_FIELDS,
    ChromiumMixin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class ChromePlugin(ChromiumMixin, BrowserPlugin):
    """Chrome browser plugin."""

    __namespace__ = "chrome"

    DIRS = (
        # Windows
        "AppData/Local/Google/Chrome/User Data/Default",
        "AppData/Local/Google/Chrome/User Data/Profile*",
        "AppData/Local/Google/Chrome/User Data/Snapshots/*/Default",
        "AppData/Local/Google/Chrome/User Data/Snapshots/*/Profile*",
        "AppData/Local/Google/Chrome/continuousUpdates/User Data/Default",
        "AppData/Local/Google/Chrome/continuousUpdates/User Data/Profile*",
        "Local Settings/Application Data/Google/Chrome/User Data/Default",
        "Local Settings/Application Data/Google/Chrome/User Data/Profile*",
        # Linux
        ".config/google-chrome/Default",
        ".config/google-chrome/Profile*",
        ".var/app/com.google.Chrome/config/google-chrome/Default",
        ".var/app/com.google.Chrome/config/google-chrome/Profile*",
        # Macos
        "Library/Application Support/Google/Chrome/Default",
        "Library/Application Support/Google/Chrome/Profile*",
    )

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chrome/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserCookieRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chrome/cookie", GENERIC_COOKIE_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chrome/download", GENERIC_DOWNLOAD_RECORD_FIELDS + CHROMIUM_DOWNLOAD_RECORD_FIELDS
    )

    BrowserExtensionRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chrome/extension", GENERIC_EXTENSION_RECORD_FIELDS
    )

    BrowserPasswordRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/chrome/password", GENERIC_PASSWORD_RECORD_FIELDS
    )

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records for Google Chrome."""
        yield from super().history("chrome")

    @export(record=BrowserCookieRecord)
    def cookies(self) -> Iterator[BrowserCookieRecord]:
        """Return browser cookie records for Google Chrome."""
        yield from super().cookies("chrome")

    @export(record=BrowserDownloadRecord)
    def downloads(self) -> Iterator[BrowserDownloadRecord]:
        """Return browser download records for Google Chrome."""
        yield from super().downloads("chrome")

    @export(record=BrowserExtensionRecord)
    def extensions(self) -> Iterator[BrowserExtensionRecord]:
        """Return browser extension records for Google Chrome."""
        yield from super().extensions("chrome")

    @export(record=BrowserPasswordRecord)
    def passwords(self) -> Iterator[BrowserPasswordRecord]:
        """Return browser password records for Google Chrome."""
        yield from super().passwords("chrome")



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.exceptions import LoaderError
from dissect.target.filesystem import VirtualFilesystem
from dissect.target.loader import open as loader_open
from dissect.target.loaders.hyperv import HyperVLoader
from dissect.target.loaders.raw import RawLoader
from dissect.target.loaders.vbk import VbkLoader
from dissect.target.loaders.vmx import VmxLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``VbkLoader`` when opening a ``Target``."""
    path = absolute_path(
        "_data/loaders/vbk/Backup Job 1/VBK-Test-VM.e56465c7-3a5a-4599-bc25-3555b9b8cD2025-07-20T160920_3702.vbk"
    )

    target = opener(path)
    assert isinstance(target._loader, VbkLoader)
    assert target.path == path


@pytest.mark.parametrize(
    ("path", "disk_sizes"),
    [
        (
            "Backup Job 1/VBK-Test-VM.e56465c7-3a5a-4599-bc25-3555b9b8cD2025-07-20T160920_3702.vbk",
            [20 * 1024 * 1024 * 1024],
        ),
        (
            "Backup Job 2_1/VBK-Test-VM-SCSI.8637b8b5-28a2-41e1-90c6-f8a0D2025-07-29T171659_FB82.vbk",
            [40 * 1024 * 1024 * 1024],
        ),
    ],
)
def test_loader(path: str, disk_sizes: list[int]) -> None:
    """Test the VBK loader on real files."""
    path = absolute_path("_data/loaders/vbk").joinpath(path)

    loader = loader_open(path)
    assert isinstance(loader, VbkLoader)

    t = Target()
    loader.map(t)
    assert len(t.disks) == len(disk_sizes)
    for disk, size in zip(sorted(t.disks, key=lambda disk: disk.size), sorted(disk_sizes), strict=False):
        assert disk.size == size


def test_mock_raw() -> None:
    """Test the VBK loader on a raw disk backup layout."""
    vfs = VirtualFilesystem()
    with patch("dissect.target.loaders.vbk.VbkFilesystem") as mock_vbk_fs:
        mock_vbk_fs.return_value = vfs

        loader = VbkLoader(Mock())
        t = Target()

        with pytest.raises(LoaderError, match="Unexpected empty VBK filesystem"):
            loader.map(t)

        vfs.makedirs("guid")

        with pytest.raises(LoaderError, match="Unsupported VBK structure"):
            loader.map(t)

        vfs.map_file_fh("guid/DEV__dummy", BytesIO("🍖👑".encode()))
        vfs.map_file_fh("guid/summary.xml", BytesIO())

        loader.map(t)
        assert isinstance(loader.loader, RawLoader)
        assert len(t.disks) == 1

        t.disks[0].seek(0)
        assert t.disks[0].read().decode() == "🍖👑"


def test_mock_vmx() -> None:
    """Test the VBK loader on a VMware backup layout."""
    vfs = VirtualFilesystem()
    with patch("dissect.target.loaders.vbk.VbkFilesystem") as mock_vbk_fs:
        mock_vbk_fs.return_value = vfs

        vfs.map_file_fh("guid/summary.xml", BytesIO())
        vfs.map_file_fh("guid/candidate.vmx", BytesIO(b'scsi0:0.fileName = "candidate.vmdk"'))
        vfs.map_file_fh(
            "guid/candidate.vmdk",
            BytesIO(b'# Disk DescriptorFile\nparentCID=ffffffff\nRW 16777216 VMFS "candidate-flat.vmdk"'),
        )
        vfs.map_file_fh("guid/candidate-flat.vmdk", BytesIO("🍖👑".encode()))

        loader = VbkLoader(Mock())
        t = Target()

        loader.map(t)
        assert isinstance(loader.loader, VmxLoader)
        assert len(t.disks) == 1

        t.disks[0].seek(0)
        assert t.disks[0].read(8).decode() == "🍖👑"


def test_mock_vmcx() -> None:
    """Test the VBK loader on a Hyper-V backup layout."""
    vfs = VirtualFilesystem()
    with (
        patch("dissect.target.loaders.vbk.VbkFilesystem") as mock_vbk_fs,
        patch("dissect.target.loaders.hyperv.hyperv.HyperVFile") as mock_hyperv_file,
    ):
        mock_vbk_fs.return_value = vfs

        mock_hyperv_file.return_value = mock_hyperv_file
        mock_hyperv_file.as_dict.return_value = {
            "configuration": {
                "manifest": {
                    "vdev0": {
                        "device": "83f8638b-8dca-4152-9eda-2ca8b33039b4",
                        "instance": "83f8638b-8dca-4152-9eda-2ca8b33039b4",
                    },
                    "vdev1": {
                        "device": "d422512d-2bf2-4752-809d-7b82b5fcb1b4",
                        "instance": "d422512d-2bf2-4752-809d-7b82b5fcb1b4",
                    },
                },
                "_83f8638b-8dca-4152-9eda-2ca8b33039b4_": {
                    "controller0": {"drive0": {"type": "VHD", "pathname": "C:\\Fake\\Path\\ide.vhdx"}},
                },
                "_d422512d-2bf2-4752-809d-7b82b5fcb1b4_": {
                    "controller0": {"drive0": {"type": "VHD", "pathname": "C:\\Fake\\Path\\scsi.vhdx"}},
                },
            }
        }

        vfs.map_file_fh("guid/summary.xml", BytesIO())
        vfs.map_file_fh("guid/Config/candidate.vmcx", BytesIO())
        vfs.map_file_fh("guid/Ide0-0/ide.vhdx", BytesIO("🐌".encode()))
        vfs.map_file_fh("guid/Scsi0-0/scsi.vhdx", BytesIO("🏃🏃".encode()))

        loader = VbkLoader(Mock())
        t = Target()

        loader.map(t)
        assert isinstance(loader.loader, HyperVLoader)
        assert len(t.disks) == 2

        disks = sorted(t.disks, key=lambda d: d.size)

        disks[0].seek(0)
        assert disks[0].read(-1).decode() == "🐌"

        disks[1].seek(0)
        assert disks[1].read(-1).decode() == "🏃🏃"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.child.virtuozzo import VirtuozzoChildTargetPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_virtuozzo(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file("/etc/vz/conf/a.conf", absolute_path("_data/plugins/child/virtuozzo/example.conf"))

    fs_unix.makedirs("vz/root/a")
    fs_unix.makedirs("vz/root/b")

    target_unix.add_plugin(VirtuozzoChildTargetPlugin)

    children = [child for _, child in target_unix.list_children()]

    assert len(children) == 2

    assert children[0].type == "virtuozzo"
    assert children[0].name == "a"
    assert str(children[0].path) == "/vz/root/a"

    assert children[1].type == "virtuozzo"
    assert str(children[1].path) == "/vz/root/b"



from __future__ import annotations

import datetime
import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

PfroRecord = TargetRecordDescriptor(
    "filesystem/windows/pfro",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("string", "operation"),
    ],
)


class PfroPlugin(Plugin):
    """PFRO plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.logfile = self.target.resolve("%windir%/PFRO.log")

    def check_compatible(self) -> None:
        if not self.logfile.exists():
            raise UnsupportedPluginError("No PFRO log found")

    @export(record=PfroRecord)
    def pfro(self) -> Iterator[PfroRecord]:
        """Return the content of %windir%/PFRO.log

        A Pending File Rename Operation log file (PFRO.log) holds information about the process of deleting or renaming
        files that are locked or being used and that will be renamed on reboot. This is related to the filerenameop
        plugin.

        References:
            - https://social.technet.microsoft.com/Forums/en-US/9b66a7b0-16d5-4d22-be4e-51df12db9f80/issue-understanding-pfro-log
            - https://community.ccleaner.com/topic/49106-pending-file-rename-operations-log/

        Yields PfroRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datetime): The parsed timestamp.
            path (uri): The parsed path.
            operation (string): The parsed operation.
        """
        target_tz = self.target.datetime.tzinfo

        try:
            for line in self.logfile.open("rt", encoding="utf-16-le"):
                if len(line) <= 1:
                    continue

                idx = line.split(" - ")
                date = idx[0]
                if "Error" in date:
                    # prfo log can log its own error. This results in an entry
                    # which gets grouped with the datetime of the logged
                    # action.
                    date = re.split(".+[A-Za-z]", date)[1]
                file_path = idx[1].split("|")[0][16:-2]
                operation = idx[1].split("|")
                operation = operation[1].split(" ")[0] if len(operation) >= 2 else None

                yield PfroRecord(
                    ts=datetime.datetime.strptime(date, "%m/%d/%Y %H:%M:%S").replace(tzinfo=target_tz),
                    path=self.target.fs.path(file_path),
                    operation=operation,
                    _target=self.target,
                )
        except UnicodeDecodeError:
            pass



from __future__ import annotations

import bz2
import dataclasses
import io
import struct
from enum import IntEnum
from functools import lru_cache
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.cstruct import BaseType, cstruct
from dissect.util.ts import wintimestamp

from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

BZIP_HEADER = b"BZh9"
FILE_MAGIC = b"ESDb"
COMPAT_1 = (1, 7, 9)


log = get_logger(__name__)


class EverythingVarInt(int, BaseType):
    @classmethod
    def _read(cls, stream: BinaryIO, context: dict[str, Any] | None = None) -> int:
        return read_varint(stream)

    @classmethod
    def _write(cls, stream: BinaryIO, data: int) -> int:
        return stream.write(write_varint(data))


c_header_def = """
flag EntryAttributes : uint32_t {
    has_file_size = 1,
    has_date_created = 2,
    has_date_modified = 4,
    has_date_accessed = 8,
    has_attributes = 16,
    has_folder_size = 32
};

struct everything_db_header {
    // Note - File header may be equal to `BZIP_HEADER`. In this case, file must be handled as a bzip compressed file
    char magic[4];

    // Version
    uint16_t version_patch;
    uint8_t version_minor;
    uint8_t version_major;

    // Flags
    EntryAttributes   entry_attributes;

    uint32_t number_of_folders;
    uint32_t number_of_files;
    EverythingVarInt number_of_filesystems;

};
"""
c_header = cstruct()
c_header.add_custom_type("EverythingVarInt", EverythingVarInt)
c_header.load(c_header_def)


def version_match(stmt: str, cond: bool) -> str:
    """Used for easy filtering of version constraints in cstruct definitions"""
    return stmt if cond else ""


@lru_cache
def filesystems_cstruct(version: tuple[int, int, int]) -> cstruct:
    c_filesystems_def = f"""
    struct EverythingVarBytes {{
        EverythingVarInt len;
        char data[len];
    }};

    struct filesystem_header {{
        EverythingVarInt type;
        {version_match("uint8_t out_of_date;", version >= COMPAT_1)}
    }};
    struct ntfs_header {{
        // Guid in format \\\\?\\Volume{{GUID}}
        EverythingVarBytes guid;
        // Disk drive (C:, D:)
        EverythingVarBytes path;
        EverythingVarBytes root;
        {version_match("EverythingVarBytes include_only;", version >= COMPAT_1)}
        {version_match("uint64_t journal_id;", version >= COMPAT_1)}
        {version_match("uint64_t next_usn;", version >= COMPAT_1)}
    }};

    struct efu_header {{
        // Path to file on disk containing EFU
        EverythingVarBytes source_path;
        {version_match("uint64_t unk1;", version >= COMPAT_1)}
    }};

    struct folder_header {{
        // Subfolder being indexed
        EverythingVarBytes path;
        // Next time to update folder
        {version_match("uint64_t next_update_time;", version >= COMPAT_1)}
    }};

    struct refs_header {{
        // Guid in format \\\\?\\Volume{{GUID}}
        EverythingVarBytes guid;
        // Disk drive (C:, D:)
        EverythingVarBytes path;
        EverythingVarBytes root;
        {version_match("EverythingVarBytes include_only;", version >= COMPAT_1)}
        {version_match("uint64_t journal_id;", version >= COMPAT_1)}
        {version_match("uint64_t next_usn;", version >= COMPAT_1)}
    }};
    """
    c_filesystems = cstruct()
    c_filesystems.add_custom_type("EverythingVarInt", EverythingVarInt)
    c_filesystems.load(c_filesystems_def)

    return c_filesystems


class EverythingFSType(IntEnum):
    NTFS = 0
    EFU = 1
    FOLDER = 2
    REFS = 3
    # TODO - Add FAT support (Only in Everything 1.5 which is still in alpha)


@dataclasses.dataclass
class Record:
    file_path: str
    size: int
    date_created: datetime | None
    date_modified: datetime | None
    date_accessed: datetime | None
    attributes: int | None
    file_type: str


class EverythingIndexObj:
    def __init__(self):
        # This is the index of filesystem_list
        self.fs_index: int | None = None
        # Path to file (Only basename)
        self.file_path = None
        # Anything not explicitly set with parent index has an fs_index
        self.parent_index = None
        self.size = None
        self.date_created = None
        self.date_modified = None
        self.date_accessed = None
        self.attributes = None

    def resolve_path(self, folder_list: list) -> str:
        if self.parent_index is not None:
            return folder_list[self.parent_index].resolve_path(folder_list) + "\\" + self.file_path
        return self.file_path

    def resolve_fs(self, folder_list: list) -> int | None:
        if self.fs_index is not None:
            return self.fs_index
        return folder_list[self.parent_index].resolve_fs(folder_list)


class EverythingDB:
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.fh.seek(0)

        # Check if file is bzipped
        header = self.fh.read(4)
        self.fh.seek(-4, io.SEEK_CUR)
        # The database may be stored in a bzip compressed file, so we check for the header.
        # Beneath the compression, the file is still a valid database file.
        if header == BZIP_HEADER:
            self.fh = bz2.open(self.fh)  # noqa: SIM115

        header = c_header.everything_db_header(self.fh)
        if header.magic != FILE_MAGIC:
            raise ValueError(f"Invalid Everything magic header {header.magic}")

        self.header = header
        self.version = (header.version_major, header.version_minor, header.version_patch)
        self.c_filesystems = filesystems_cstruct(self.version)
        self.filesystems = []
        self.__parse_filesystems()

        # This might be hidden/system files/folders, or maybe a list of folders to exclude?
        # TODO - Check what happens when changing exclude hidden/system files/folders
        if self.version > COMPAT_1:
            exclude_flags = c_header.uint8_t(self.fh)
            log.debug("Exclude flags: %s", exclude_flags)

        # There is some logic here, but in my test file it doesn't affect the data
        # (Tested by dumping the raw array while debugging in IDA)
        # In practice, none of the data in the function is saved, it just advances the offset
        # This *MIGHT* be affected by an old version of the db (before 1.7.8)
        # If any of these fail, then I'd have to reverse this function and add support

        # I expect 0 here because the inner implementation of the function is:
        # for i in range(read_varint(self.fh)):
        #   do something
        # and the function is called 3 times one after another. As long as zero is returned each time, we don't need
        # to implement the logic
        if [read_varint(self.fh), read_varint(self.fh), read_varint(self.fh)] != [0, 0, 0]:
            raise NotImplementedError("Failed to parse database, unimplemented feature. Please open an issue")

        # Keep current position in the file, so we can seek back to it during __iter__
        self._start_seek = self.fh.tell()

    def __repr__(self) -> str:
        return f"{self.header} - Filesystems{self.filesystems}"

    def __parse_filesystems(self) -> None:
        for i in range(self.header.number_of_filesystems):
            filesystem_header = self.c_filesystems.filesystem_header(self.fh)
            log.debug("Filesystem %d: type %s", i, filesystem_header)

            try:
                fs_type = EverythingFSType(filesystem_header.type)
            except ValueError:
                raise ValueError(f"Unknown FS type {filesystem_header.type}")

            if fs_type in [EverythingFSType.NTFS, EverythingFSType.REFS]:
                if fs_type == EverythingFSType.NTFS:
                    header = self.c_filesystems.ntfs_header(self.fh)
                else:
                    header = self.c_filesystems.refs_header(self.fh)

            elif fs_type == EverythingFSType.EFU:
                header = self.c_filesystems.efu_header(self.fh)

            elif fs_type == EverythingFSType.FOLDER:
                header = self.c_filesystems.folder_header(self.fh)
            else:
                raise NotImplementedError(f"Have not implemented parsing {fs_type}")
            self.filesystems.append((fs_type, filesystem_header, header))

    def __iter__(self) -> Iterator[Record]:
        self.fh.seek(self._start_seek)
        # This builds a list of folders in the filesystem.  This creates an index, where each object contains:
        #   index of parent object (meaning the folder above)
        #   index of filesystem
        # This is later used to build a hierarchy for folders
        folder_list = [EverythingIndexObj() for _ in range(self.header.number_of_folders)]
        parent_lookup = c_header.uint32_t[self.header.number_of_folders](self.fh)
        for lookup_index, folder in enumerate(folder_list):
            parent_index = parent_lookup[lookup_index]
            # `parent_index` is an index into `folder_list`, which points to the parent folder of the current item
            # At the end of the loop, we have a list where each folder has an index to its parent.
            # Root folders (Such as C:\) don't have parents, instead they have an index into self.filesystem_list,
            # representing the filesystem they belong to.
            # This is represented by `filesystem_index = parent_index - self.header.number_of_folders`.
            # So we identify root folders by their parent_index being bigger than self.header.number_of_folders.

            # If `parent_index` is too big, we have an issue (It can't be an index into a filesystem or parent)
            if parent_index >= (self.header.number_of_filesystems + self.header.number_of_folders):
                raise ValueError("Invalid folder offset")
            if parent_index >= self.header.number_of_folders:
                folder.fs_index = parent_index - self.header.number_of_folders
            else:
                folder.parent_index = parent_index

        # This recursively resolves fs_index, so every folder will have it.
        for folder in folder_list:
            folder.fs_index = folder.resolve_fs(folder_list)

        yield from parse_folders(self, folder_list)

        yield from parse_files(self, folder_list)


def read_varint(stream: BinaryIO) -> int:
    """Read an ``uint8``, if it's equal to ``0xFF``, read the next 4 bytes as an ``int32``.
    In decompiled-ish code:
    .. code-block:: c

        int v1;
        LOBYTE(v1) = read(fd, 1);
        if ( (_BYTE)v1 == 0xFF ) v1 = read(fd, 4);
        else v1 = (unsigned __int8)v1;
    """
    first = stream.read(1)[0]
    if first == 0xFF:
        return int.from_bytes(stream.read(4), byteorder="little", signed=True)
    return first


def write_varint(data: int) -> bytes:
    # Strings under 0xFF are stored as a single unsigned byte, and everything else is stored as 0xFF + 4 bytes signed
    if data < 0xFF:
        return struct.pack("<B", data)
    return struct.pack("<Bi", 0xFF, data)


def parse_folder(db: EverythingDB, folder: EverythingIndexObj, name: str) -> None:
    folder.file_path = name
    # This is hardcoded for all folders
    folder.attributes = 16

    if c_header.EntryAttributes.has_folder_size in db.header.entry_attributes:
        folder.size = c_header.uint64_t(db.fh)
    if c_header.EntryAttributes.has_date_created in db.header.entry_attributes:
        folder.date_created = c_header.uint64_t(db.fh)
    if c_header.EntryAttributes.has_date_modified in db.header.entry_attributes:
        folder.date_modified = c_header.uint64_t(db.fh)
    if c_header.EntryAttributes.has_date_accessed in db.header.entry_attributes:
        folder.date_accessed = c_header.uint64_t(db.fh)
    if c_header.EntryAttributes.has_attributes in db.header.entry_attributes:
        folder.attributes = c_header.uint32_t(db.fh)

    if db.filesystems[folder.fs_index][0] == EverythingFSType.REFS:
        # Unknown
        c_header.uint64_t(db.fh)
        c_header.uint64_t(db.fh)
    elif db.filesystems[folder.fs_index][0] == EverythingFSType.NTFS:
        # Unknown
        c_header.uint64_t(db.fh)
    elif db.filesystems[folder.fs_index][0] == EverythingFSType.EFU:
        if folder.parent_index is None:
            # The EFU format does not contain the root drive, so it just puts random data into
            # the metadata.  This will cause errors if passed to flow.record, so we remove it here
            folder.date_accessed = None
            folder.date_modified = None
            folder.date_created = None
            folder.size = None

    # Must be done here because EFU files can contain garbage values
    if folder.date_accessed is not None:
        folder.date_accessed = wintimestamp(folder.date_accessed)
    if folder.date_modified is not None:
        folder.date_modified = wintimestamp(folder.date_modified)
    if folder.date_created is not None:
        folder.date_created = wintimestamp(folder.date_created)


def read_truncated_name(fh: BinaryIO, current_buf: bytes = b"") -> bytes:
    """Read a string stored in the format used by the database.
    If you have called this function before, you *must* pass the previous result to `current_buf`.

    Explanation:
    Everything has an "Optimization", where it saves all the basenames of the folders (and files)
    to the disk alphabetically.  This allows them to reuse similar filename buffers.
    For example, if two folders in the filesystem are named "AAA" and "ABCD",
    (and are alphabetically consecutive)
    then the first file will have data "AAA", with a `new_byte_count` of 3,
    and the second file will have a `new_byte_count` of 3 (length of "BCD"),
    and a `trunc_from_prev` of 3,
    thereby telling us to remove the last 3 bytes of the previous buffer, and saving space.
    The same thing happens later on when parsing filenames
    """
    if new_byte_count := read_varint(fh):
        trunc_from_prev = read_varint(fh)
        if trunc_from_prev > len(current_buf):
            raise ValueError(f"Error while parsing folder names {trunc_from_prev} > {len(current_buf)}")
        current_buf = current_buf[: len(current_buf) - trunc_from_prev]
    current_buf += fh.read(new_byte_count)
    return current_buf


def parse_folders(db: EverythingDB, folder_list: list[EverythingIndexObj]) -> Iterator[Record]:
    temp_buf = b""
    for folder in folder_list:
        # There is a bug loading EFU files, (which causes Everything v1.4.0.704b to crash)
        # where it tries to read an extra byte, which is not present in the file.
        # If at any point parsing EFU files fails, it is likely because of this bug,
        # and can be fixed by reading and discarding a single byte here.

        temp_buf = read_truncated_name(db.fh, temp_buf)
        # The yield can't happen here, because we can only call resolve_path once we finish this loop
        parse_folder(db, folder, temp_buf.decode())

    for folder in folder_list:
        yield Record(
            file_path=folder.resolve_path(folder_list),
            size=folder.size,
            attributes=folder.attributes,
            date_created=folder.date_created,
            date_modified=folder.date_modified,
            date_accessed=folder.date_accessed,
            file_type="directory",
        )


def parse_files(db: EverythingDB, folder_list: list[EverythingIndexObj]) -> Iterator[Record]:
    temp_buf = b""
    for _ in range(db.header.number_of_files):
        parent_index = c_header.uint32_t(db.fh)
        if parent_index > db.header.number_of_filesystems + db.header.number_of_folders:
            raise ValueError(
                "Error while parsing file names. Parent index out of bounds "
                f"{parent_index} > {db.header.number_of_filesystems + db.header.number_of_folders}"
            )

        # There MAY be some edge case where this is okay, and the check should be
        # parent_index < db.number_of_folders + db.total_filesystem_num, but I haven't seen this yet
        # Is there any way a file has a root filesystem without a folder?
        if parent_index >= db.header.number_of_folders:
            raise ValueError("Something weird, this points to filesystem_index")

        # There is a bug loading EFU files, (which causes Everything v1.4.0.704b to crash)
        # where it tries to read an extra byte, which is not present in the file.
        # If at any point parsing EFU files fails, it is likely because of this bug,
        # and can be fixed by reading and discarding a single byte here.

        file_name = folder_list[parent_index].resolve_path(folder_list)
        temp_buf = read_truncated_name(db.fh, temp_buf)
        file_size = (
            c_header.uint64_t(db.fh) if db.header.entry_attributes.has_file_size in db.header.entry_attributes else None
        )
        date_created = (
            wintimestamp(c_header.uint64_t(db.fh))
            if db.header.entry_attributes.has_date_created in db.header.entry_attributes
            else None
        )
        date_modified = (
            wintimestamp(c_header.uint64_t(db.fh))
            if db.header.entry_attributes.has_date_modified in db.header.entry_attributes
            else None
        )
        date_accessed = (
            wintimestamp(c_header.uint64_t(db.fh))
            if db.header.entry_attributes.has_date_accessed in db.header.entry_attributes
            else None
        )
        attributes = (
            c_header.uint32_t(db.fh)
            if db.header.entry_attributes.has_attributes in db.header.entry_attributes
            else None
        )

        try:
            yield Record(
                file_path=f"{file_name}\\{temp_buf.decode()}",
                size=file_size,
                attributes=attributes,
                date_created=date_created,
                date_modified=date_modified,
                date_accessed=date_accessed,
                file_type="file",
            )
        # This shouldn't be possible, but it happened in my tests to folders in the recycle bin
        except UnicodeDecodeError as e:
            log.warning("Failed parsing filepath: %s\\%s", file_name, temp_buf, exc_info=e)



from __future__ import annotations

import os
import urllib.parse
from pathlib import Path
from typing import TYPE_CHECKING, Generic, TypeVar

from dissect.target.helpers.lazy import import_lazy
from dissect.target.helpers.loaderutil import parse_path_uri
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

__all__ = [
    "Loader",
    "RawLoader",
    "open",
    "register",
]

log = get_logger(__name__)

DirLoader: Loader = import_lazy("dissect.target.loaders.dir").DirLoader
"""A lazy loaded :class:`dissect.target.loaders.dir.DirLoader`."""

RawLoader: Loader = import_lazy("dissect.target.loaders.raw").RawLoader
"""A lazy loaded :class:`dissect.target.loaders.raw.RawLoader`."""

LOADERS: list[Loader] = []
LOADERS_BY_SCHEME: dict[str, Loader] = {
    "dir": DirLoader,
    "raw": RawLoader,
}
MODULE_PATH = "dissect.target.loaders"


class Loader:
    """A base class for loading a specific path and coupling it to a :class:`Target <dissect.target.target.Target>`.

    Implementors of this class are responsible for mapping any type of source data
    to a :class:`Target <dissect.target.target.Target>`.
    Whether that's to map all VMDK files from a VMX or mapping the contents of a zip file to a virtual filesystem,
    if it's something that can be translated to a "disk", "volume" or "filesystem", you can write a loader that
    maps it into a target.

    You can do anything you want to manipulate the :class:`Target <dissect.target.target.Target>` object
    in your ``map`` function, but generally you do one of the following:

    * open a :class:`Container <dissect.target.container.Container>` and add it to ``target.disks``.
    * open a :class:`Volume <dissect.target.volume.Volume>` and add it to ``target.volumes``.
    * open a :class:`VirtualFilesystem <dissect.target.filesystem.VirtualFilesystem>`,\
    add your files into it and add it to ``target.filesystems``.

    You don't need to manually parse volumes or filesystems in your loader, just add the highest level object you have
    (e.g. a :class:`Container <dissect.target.container.Container>` of a VMDK file) to the target.
    However, sometimes you need to get creative.
    Take a look at the :class:`ITunesLoader <dissect.target.loaders.itunes.ITunesLoader>` and
    :class:`TarLoader <dissect.target.loaders.tar.TarLoader>` for some creative examples.

    Args:
        path: The target path to load.
        parsed_path: A URI parsed path to use.
    """

    def __init__(
        self, path: Path, *, parsed_path: urllib.parse.ParseResult | None = None, resolve: bool = True, **kwargs
    ):
        self.path = path
        self.absolute_path = None
        if resolve:
            try:
                self.absolute_path = path.resolve()
            except Exception:
                log.debug("Failed to resolve loader path %r", path)
                self.absolute_path = path
            self.base_path = self.absolute_path.parent
        self.parsed_path = parsed_path
        self.parsed_query = (
            dict(urllib.parse.parse_qsl(parsed_path.query, keep_blank_values=True)) if parsed_path else {}
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self.path)!r})"

    @staticmethod
    def detect(path: Path) -> bool:
        """Detects wether this ``Loader`` class can load this specific ``path``.

        Args:
            path: The target path to check.

        Returns:
            ``True`` if the ``path`` can be loaded by a ``Loader`` instance. ``False`` otherwise.
        """
        raise NotImplementedError

    @staticmethod
    def find_all(path: Path, parsed_path: urllib.parse.ParseResult | None = None) -> Iterator[str | Path]:
        """Finds all targets to load from ``path``.

        This can be used to open multiple targets from a target path that doesn't necessarily map to files on a disk.
        For example, a wildcard in a hostname a loader that opens targets from an API or Unix socket,
        such as the Carbon Black loader.

        Args:
            path: The location to a target to try and open multiple paths from.
            parsed_path: A URI parsed path to use.

        Returns:
            All the target paths found from the source path.
        """
        yield path

    def map(self, target: Target) -> None:
        """Maps the loaded path into a ``Target``.

        Args:
            target: The target that we're mapping into.
        """
        raise NotImplementedError


T = TypeVar("T")


class SubLoader(Loader, Generic[T]):
    """A base class for loading arbitary data and coupling it to a :class:`Target <dissect.target.target.Target>`.

    Should not be called like a regular :class:`Loader`. For examples see :class:`TarLoader`
    and :class:`TarSubLoader` implementations.
    """

    def __init__(
        self,
        path: Path,
        value: T,
        *,
        parsed_path: urllib.parse.ParseResult | None = None,
        resolve: bool = True,
        **kwargs,
    ):
        super().__init__(path, parsed_path=parsed_path, resolve=resolve, **kwargs)

    @staticmethod
    def detect(path: Path, value: T) -> bool:
        raise NotImplementedError

    def map(self, target: Target) -> None:
        raise NotImplementedError


def register(module_name: str, class_name: str, internal: bool = True) -> None:
    """Registers a ``Loader`` class inside ``LOADERS``.

    This function registers a loader using ``modname`` relative to the ``MODULE_PATH``.
    It lazily imports the module, and retrieves the specific class from it.

    Args:
        module: The module where to find the loader.
        class_name: The class to load.
        internal: Whether it is an internal module or not.
    """
    module = f"{MODULE_PATH}.{module_name}" if internal else module_name

    loader = getattr(import_lazy(module), class_name)
    LOADERS.append(loader)
    LOADERS_BY_SCHEME[module_name] = loader


def find_loader(
    path: str | Path,
    *,
    fallbacks: list[type[Loader]] | None = None,
) -> type[Loader] | None:
    """Finds a :class:`Loader` class for a given path.

    This searches for a specific :class:`Loader` classs that is able to load a target pointed to by ``path``.
    Once it detects a suitable :class:`Loader` it immediately returns said :class:`Loader` class.

    The :class:`DirLoader <dissect.target.loaders.dir.DirLoader>` is used as the last entry
    due to how the detection methods function.

    Args:
        path: The target path to load.
        fallbacks: Fallback loaders to try.

    Returns:
        A :class:`Loader` class for the specific target if one is found.
    """
    path = Path(path) if not isinstance(path, os.PathLike) else path

    if fallbacks is None:
        fallbacks = [DirLoader]

    for loader in LOADERS + fallbacks:
        try:
            if loader.detect(path):
                return loader
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", loader)
            log.debug("", exc_info=e)

    return None


def find_loader_by_scheme(scheme: str) -> type[Loader] | None:
    """Finds a :class:`Loader` class by its scheme.

    Args:
        scheme: The scheme to find the loader for.

    Returns:
        A :class:`Loader` class for the specific scheme if one is found.
    """
    return LOADERS_BY_SCHEME.get(scheme)


def open(path: str | Path, *, fallbacks: list[type[Loader]] | None = None, **kwargs) -> Loader | None:
    """Opens a :class:`Loader` for a given path.

    This instantiates a :class:`Loader` for a specific ``path``.
    The :class:`DirLoader <dissect.target.loaders.dir.DirLoader>` is used as the last entry
    due to how the detection methods function.

    Args:
        path: The target path to load.

    Returns:
        A :class:`Loader` instance for the specific target if one is found.
    """
    adjusted_path, parsed_path = parse_path_uri(path)

    if parsed_path is not None and (loader := find_loader_by_scheme(parsed_path.scheme)):
        # If we find a loader by scheme, use the adjusted path
        kwargs["parsed_path"] = parsed_path
        return loader(adjusted_path, **kwargs)

    # Use the original unadjusted path
    path = Path(path) if not isinstance(path, os.PathLike) else path
    if loader := find_loader(path, fallbacks=fallbacks):
        kwargs["parsed_path"] = parsed_path
        return loader(path, **kwargs)

    return None


register("local", "LocalLoader")
register("ad1", "AD1Loader")
register("asdf", "AsdfLoader")
register("vmx", "VmxLoader")
register("vmwarevm", "VmwarevmLoader")
register("hyperv", "HyperVLoader")
register("pvs", "PvsLoader")
register("pvm", "PvmLoader")
register("utm", "UtmLoader")
register("libvirt", "LibvirtLoader")
register("proxmox", "ProxmoxLoader")
register("ovf", "OvfLoader")
register("ova", "OvaLoader")
register("vbox", "VBoxLoader")
register("vb", "VBLoader")
register("vbk", "VbkLoader")
register("xva", "XvaLoader")
register("vma", "VmaLoader")
register("kape", "KapeLoader")
register("velociraptor", "VelociraptorLoader")
register("tanium", "TaniumLoader")
register("itunes", "ITunesLoader")
register("ab", "AndroidBackupLoader")
register("cellebrite", "CellebriteLoader")
register("target", "TargetLoader")
# Disabling ResLoader because of DIS-536
# register("res", "ResLoader")
register("overlay2", "Overlay2Loader")
register("overlay", "OverlayLoader")
register("phobos", "PhobosLoader")
register("uac", "UacLoader")
register("vmsupport", "VmSupportLoader")
# tar and zip loaders should be low priority to give other loaders a chance first
register("tar", "TarLoader")
register("zip", "ZipLoader")
# These are URI based loaders, so just put them low down the list since they don't do any detection anyway
register("smb", "SmbLoader")
register("cb", "CbLoader")
register("cyber", "CyberLoader")
register("log", "LogLoader")
register("remote", "RemoteLoader")
register("mqtt", "MqttLoader")
register("multiraw", "MultiRawLoader")  # Should be last



from __future__ import annotations

import gzip
from typing import TYPE_CHECKING

from dissect.target.filesystems.apfs import ApfsFilesystem
from dissect.target.helpers import keychain
from tests._utils import absolute_path

if TYPE_CHECKING:
    import pytest


def test_apfs_encrypted(caplog: pytest.LogCaptureFixture) -> None:
    """Test that encrypted APFS filesystems are correctly detected."""
    with gzip.open(absolute_path("_data/filesystems/apfs/encrypted.bin.gz"), "rb") as fh:
        assert ApfsFilesystem.detect(fh)

        fs = ApfsFilesystem(fh)
        assert caplog.messages == [
            "Failed to open APFS volume 'Encrypted' (0a51df00-e6ff-6949-9607-efa24d864392): No valid decryption key found"  # noqa: E501
        ]

        # Test with wildcard key
        keychain.register_wildcard_value("password")
        fs = ApfsFilesystem(fh)
        assert len(list(fs.iter_subfs())) == 1
        assert len(next(fs.iter_subfs()).get("/").listdir()) == 11

        # Test with UUID
        keychain.KEYCHAIN.clear()
        keychain.register_key(
            keychain.KeyType.PASSPHRASE,
            "password",
            "0a51df00-e6ff-6949-9607-efa24d864392",
            provider="apfs",
        )
        fs = ApfsFilesystem(fh)
        assert len(list(fs.iter_subfs())) == 1
        assert len(next(fs.iter_subfs()).get("/").listdir()) == 11


def test_apfs_direntry() -> None:
    """Test APFS directory entries and symlink behavior."""
    with gzip.open(absolute_path("_data/filesystems/apfs/encrypted.bin.gz"), "rb") as fh:
        assert ApfsFilesystem.detect(fh)

        keychain.register_wildcard_value("password")
        fs = ApfsFilesystem(fh)
        volume = next(fs.iter_subfs())
        root = volume.get("/")

        dirents = {entry.name: entry for entry in root.scandir()}
        assert len(dirents) == 11

        assert dirents["dir"].is_dir(follow_symlinks=False)
        assert dirents["dir"].is_dir(follow_symlinks=True)
        assert dirents["symlink-dir"].is_symlink()
        assert not dirents["symlink-dir"].is_dir(follow_symlinks=False)
        assert dirents["symlink-dir"].is_dir(follow_symlinks=True)



from __future__ import annotations

import stat
from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.extfs import extfs

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import DirEntry, Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator


class ExtFilesystem(Filesystem):
    __type__ = "ext"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, **kwargs)
        self.extfs = extfs.ExtFS(fh)

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        fh.seek(1024)
        return fh.read(512)[56:58] == b"\x53\xef"

    def get(self, path: str) -> FilesystemEntry:
        return ExtFilesystemEntry(self, path, self._get_node(path))

    def _get_node(self, path: str, node: extfs.INode | None = None) -> extfs.INode:
        try:
            return self.extfs.get(path, node)
        except extfs.FileNotFoundError as e:
            raise FileNotFoundError(path) from e
        except extfs.NotADirectoryError as e:
            raise NotADirectoryError(path) from e
        except extfs.NotASymlinkError as e:
            raise NotASymlinkError(path) from e
        except extfs.Error as e:
            raise FileNotFoundError(path) from e


class ExtDirEntry(DirEntry):
    fs: ExtFilesystem
    entry: extfs.INode

    def get(self) -> ExtFilesystemEntry:
        return ExtFilesystemEntry(self.fs, self.path, self.entry)

    def stat(self, *, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.get().stat(follow_symlinks=follow_symlinks)


class ExtFilesystemEntry(FilesystemEntry):
    fs: ExtFilesystem
    entry: extfs.INode

    def get(self, path: str) -> FilesystemEntry:
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return ExtFilesystemEntry(self.fs, full_path, self.fs._get_node(path, self.entry))

    def open(self) -> BinaryIO:
        if self.is_dir():
            raise IsADirectoryError(self.path)
        return self._resolve().entry.open()

    def scandir(self) -> Iterator[ExtDirEntry]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for fname, f in self._resolve().entry.listdir().items():
            if fname in (".", ".."):
                continue

            # TODO: Separate INodes and DirEntry in dissect.extfs
            yield ExtDirEntry(self.fs, self.path, fname, f)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.filetype == stat.S_IFDIR
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks).entry.filetype == stat.S_IFREG
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return self.entry.filetype == stat.S_IFLNK

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError

        return self.entry.link

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        node = self.entry.inode

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = fsutil.stat_result(
            [
                node.i_mode,
                self.entry.inum,
                id(self.fs),
                node.i_links_count,
                node.i_uid,
                node.i_gid,
                self.entry.size,
                # timestamp() returns a float which will fill both the integer and float fields
                self.entry.atime.timestamp(),
                self.entry.mtime.timestamp(),
                self.entry.ctime.timestamp(),
            ]
        )

        # Set birthtime if available
        if self.entry.crtime:
            st_info.st_birthtime = self.entry.crtime.timestamp()
            st_info.st_birthtime_ns = self.entry.crtime_ns

        # Set the nanosecond resolution separately
        st_info.st_atime_ns = self.entry.atime_ns
        st_info.st_mtime_ns = self.entry.mtime_ns
        st_info.st_ctime_ns = self.entry.ctime_ns

        # Set blocks
        st_info.st_blocks = self.entry.inode.i_blocks_lo
        st_info.st_blksize = self.entry.extfs.block_size

        return st_info

    def attr(self) -> Any:
        return self._resolve().entry.xattr

    def lattr(self) -> Any:
        return self.entry.xattr



from __future__ import annotations

import itertools
import platform
import stat
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from dissect.target.filesystem import VirtualFile, VirtualFilesystem
from dissect.target.helpers import fsutil
from dissect.target.loaders.tar import TarLoader
from dissect.target.plugins.filesystem.walkfs import WalkFsPlugin
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.target import Target


def test_walkfs_plugin(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test basic walkfs plugin behavior."""

    fs_unix.map_file_fh("/path/to/some/file", BytesIO(bytes.fromhex("89504e470d0a1a0a")))
    fs_unix.map_file_fh("/path/to/some/other/file.ext", BytesIO(b""))
    fs_unix.map_file_fh("/root_file", BytesIO(b""))
    fs_unix.map_file_fh("/other_root_file.ext", BytesIO(b""))
    fs_unix.map_file_fh("/.test/test.txt", BytesIO(b""))
    fs_unix.map_file_fh("/.test/.more.test.txt", BytesIO(b""))
    fs_unix.symlink("/.test/.more.test.txt", "/.test/.more.test.symlink.txt")

    target_unix.add_plugin(WalkFsPlugin)

    results = sorted(target_unix.walkfs(mimetype=True), key=lambda r: r.path)
    assert len(results) == 15
    assert [r.path for r in results] == [
        "/",
        "/.test",
        "/.test/.more.test.symlink.txt",
        "/.test/.more.test.txt",
        "/.test/test.txt",
        "/etc",
        "/other_root_file.ext",
        "/path",
        "/path/to",
        "/path/to/some",
        "/path/to/some/file",
        "/path/to/some/other",
        "/path/to/some/other/file.ext",
        "/root_file",
        "/var",
    ]
    assert results[0].type == "dir"
    assert results[2].type == "symlink"
    assert results[3].type == "file"

    separator = "\\" if platform.system() == "Windows" else ""
    assert {fsutil.normalize(str(r.path), alt_separator=separator): r.mimetype for r in results if r.mimetype} == {
        "/.test/.more.test.symlink.txt": "text/plain",  # inferred by txt extension
        "/.test/.more.test.txt": "text/plain",  # inferred by txt extension
        "/.test/test.txt": "text/plain",  # inferred by txt extension
        "/path/to/some/file": "image/png",  # inferred by magic bytes
    }


@pytest.mark.benchmark
@pytest.mark.parametrize(
    "params",
    [
        pytest.param({}, id="default"),
        pytest.param({"mimetype": True}, id="mimetype"),
    ],
)
def test_benchmark_walkfs(target_bare: Target, benchmark: BenchmarkFixture, params: dict) -> None:
    """Benchmark walkfs performance on a small tar archive with ~500 files."""

    loader = TarLoader(Path(absolute_path("_data/loaders/containerimage/alpine-docker.tar")))
    loader.map(target_bare)
    target_bare.apply()

    benchmark(lambda: list(itertools.islice(WalkFsPlugin(target_bare).walkfs(**params), 100)))


def test_walkfs_suid(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect a SUID binary correctly in the WalkFS plugin."""

    vfile = VirtualFile(fs_unix, "binary", None)
    vfile.lstat = Mock()
    vfile.lstat.return_value = fsutil.stat_result([stat.S_IFREG | stat.S_ISUID, 0, 0, 0, 0, 0, 0, 0, 0, 0])
    fs_unix.map_file_entry("/path/to/suid/binary", vfile)

    target_unix.add_plugin(WalkFsPlugin)

    results = list(target_unix.walkfs())
    assert len(results) == 7

    assert results[-1].path == "/path/to/suid/binary"
    assert results[-1].fs_types == ["virtual"]
    assert results[-1].mode == 34816
    assert results[-1].is_suid


def test_walkfs_xattr(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we parse xattrs correctly in the WalkFS plugin."""

    xattr1 = Mock()
    xattr1.name = "security.capability"
    xattr1.value = bytes.fromhex("010000010020f00000f00f0f")

    xattr2 = Mock()
    xattr2.name = "example.attr"
    xattr2.value = b"some value"

    vfile1 = VirtualFile(fs_unix, "file", None)
    vfile1.lattr = Mock()
    vfile1.lattr.return_value = [xattr1, xattr2]
    vfile1.fs.__type__ = "extfs"
    fs_unix.map_file_entry("/path/to/xattr1/file", vfile1)

    target_unix.add_plugin(WalkFsPlugin)

    results = list(target_unix.walkfs(capability=True))
    assert len(results) == 8

    assert results[-2].path == "/path/to/xattr1/file"
    assert results[-2].fs_types == ["extfs"]
    assert results[-2].attr == ["security.capability=010000010020f00000f00f0f", "example.attr=736f6d652076616c7565"]

    assert results[-1].mtime
    assert results[-1].permitted == ["CAP_NET_RAW", "CAP_SYS_PACCT", "CAP_SYS_ADMIN", "CAP_SYS_BOOT", "CAP_SYS_NICE"]
    assert results[-1].inheritable == [
        "CAP_NET_ADMIN",
        "CAP_NET_RAW",
        "CAP_IPC_LOCK",
        "CAP_IPC_OWNER",
        "CAP_SYS_MODULE",
        "CAP_SYS_RAWIO",
        "CAP_SYS_CHROOT",
        "CAP_SYS_PTRACE",
        "CAP_SYS_RESOURCE",
        "CAP_SYS_TIME",
        "CAP_SYS_TTY_CONFIG",
        "CAP_MKNOD",
    ]
    assert results[-1].effective
    assert results[-1].root_id is None



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.record import EmptyRecord
from dissect.target.plugin import OSPlugin, export, internal

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record
    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class DefaultOSPlugin(OSPlugin):
    def __init__(self, target: Target):
        super().__init__(target)
        if len(target.filesystems) == 1:
            target.fs.mount("/", target.filesystems[0])
        else:
            for i, fs in enumerate(target.filesystems):
                target.fs.mount(f"fs{i}", fs)

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        pass

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        if sysvol:
            target.fs.mount("/", sysvol)
        return cls(target)

    @export(property=True)
    def hostname(self) -> str | None:
        pass

    @export(property=True)
    def ips(self) -> list[str]:
        return []

    @export(property=True)
    def version(self) -> str | None:
        pass

    @export(record=EmptyRecord)
    def users(self) -> Iterator[Record]:
        yield from ()

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        yield from ()

    @export(property=True)
    def os(self) -> str:
        return "default"

    @export(property=True)
    def architecture(self) -> str | None:
        pass



from __future__ import annotations

import re
from collections import defaultdict
from configparser import ConfigParser, MissingSectionHeaderError
from io import StringIO
from itertools import chain
from re import Match, compile, sub
from typing import TYPE_CHECKING, Any

from defusedxml import ElementTree

from dissect.target.exceptions import PluginError
from dissect.target.helpers import configutil
from dissect.target.helpers.logging import get_logger

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target


log = get_logger(__name__)

try:
    from ruamel.yaml import YAML

    HAS_YAML = True
except ImportError:
    HAS_YAML = False

IGNORED_IPS = [
    "0.0.0.0",
    "127.0.0.1",
    "::1",
    "0:0:0:0:0:0:0:1",
    "0:0:0:0:0:0:0:0",
]


class Template:
    """Class that represents a parsing template. Linux network configuration files can be parsed according to the
    options specified within this template.

    Args:
        name: Name of the parsing template.
        parser: A function or object to parse a Linux network configuration file with.
        sections: Configuration sections to look for in the specified configuration file.
        options: Configuration options to look for in the specified configuration file (ip, dns, dhcp, etc.).
    """

    def __init__(self, name: str, parser: Any, sections: list[str], options: list[str]):
        self.name = name
        self.parser = parser

        self.sections = sections
        self.options = options

    def set_name(self, name: str) -> None:
        """Sets the name of the the used parsing template to the name of the discovered network manager."""
        self.name = name

    def create_config(self, path: TargetPath) -> dict | None:
        """Create a network config dictionary based on the configured template and supplied path.

        Args:
            Path: Path to the to be parsed config file.

        Returns:
            A dictionary based on the provided configured template, else None
        """

        if not path.exists() or path.is_dir():
            log.debug("Failed to get config file %s", path)
            config = None

        if self.name == "netplan":
            config = self._parse_netplan_config(path)
        elif self.name == "wicked":
            config = self._parse_wicked_config(path)
        elif self.name == "interfaces":
            config = self._parse_text_config(("#"), " ", path)
        elif isinstance(self.parser, ConfigParser):
            config = self._parse_configparser_config(path)
        return config

    def _parse_netplan_config(self, path: TargetPath) -> dict | None:
        """Internal function to parse a netplan YAML based configuration file into a dict.

        Args:
            fh: A path to the configuration file to be parsed.

        Returns:
            Dictionary containing the parsed YAML based configuration file.
        """
        if HAS_YAML:
            return self.parser(path.open("rb"))
        log.error("Failed to parse %s. Cannot import ruamel.yaml", self.name)
        return None

    def _parse_wicked_config(self, path: TargetPath) -> dict:
        """Internal function to parse a wicked XML based configuration file into a dict.

        Args:
            fh: A path to the configuration file to be parsed.

        Returns:
            Dictionary containing the parsed xml based Linux network manager based configuration file.
        """
        # nasty workaround for namespaced XML without namespace (xlmns) definitions
        # we have to replace the ":" for this with "___" (three underscores) to make the xml config non-namespaced.
        pattern = compile(r"(?<=\n)\s+(<.+?>)")

        def replace_match(match: Match) -> str:
            return match.group(1).replace(":", "___")

        text = sub(pattern, replace_match, path.open("rt").read())

        xml = self.parser.parse(StringIO(text))
        return self._parse_xml_config(xml, self.sections, self.options)

    def _parse_configparser_config(self, path: TargetPath) -> dict:
        """Internal function to parse ConfigParser compatible configuration files into a dict.

        Args:
            path: A path to the configuration file to be parsed.

        Returns:
            Dictionary containing the parsed ConfigParser compatible configuration file.
        """
        try:
            self.parser.read_string(path.open("rt").read(), path.name)
        except MissingSectionHeaderError:
            # configparser does like config files without headers, so we inject a header to make it work.
            self.parser.read_string(f"[{self.name}]\n" + path.open("rt").read(), path.name)

        return self.parser._sections

    def _parse_text_config(self, comments: str, delim: str, path: TargetPath) -> dict:
        """Internal function to parse a basic plain text based configuration file into a dict.

        Args:
            comments: A string value defining the comment style of the configuration file.
            delim: A string value defining the delimiters used in the configuration file.
            path: A path to the configuration file to be parsed.

        Returns:
            Dictionary with a parsed plain text based Linux network manager configuration file.
        """
        config = defaultdict(dict)
        option_dict = {}

        for line in path.open("rt"):
            line = line.strip()

            if not line or line.startswith(comments):
                continue

            key, value = line.split(delim, maxsplit=1)
            key = key.strip()
            value = value.strip()

            for option in self.options:
                if option == key:
                    # This key may have multiple values
                    if key == "dns-nameservers":
                        value = value.split(delim)

                    if option_dict.get(key):
                        option_dict[key].append(value) if isinstance(value, str) else option_dict[key].extend(value)
                    else:
                        option_dict[key] = [value] if isinstance(value, str) else value

        for section in self.sections:
            config[section] = option_dict

        return dict(config)

    def _parse_xml_config(self, xml: ElementTree, sections: list, options: list) -> dict:
        """Internal function to parse a xml based Linux network manager based configuration file into a dict.

        Args:
            xml: An XML ElementTree object to convert to a dict
            sections: Configuration sections to look-up in the specified XML ElementTree
            options: Configuration options to look-up in the specified XML ElementTree

        Returns:
            Dictionary containing the parsed xml based Linux network manager based configuration file.
        """
        xml_dict = {}
        for section in sections:
            tag = section
            # as a side-effect of escaping the ":" value we have to escape values containing ":" here as well.
            section = section.replace(":", "___")
            for element in (element for element in xml.findall(".//") if element.tag == section):
                text = element.text.strip()
                # a non-empty text field indicates that we are done searching for values
                if text:
                    xml_dict[tag] = {tag: text}
                else:
                    for option, value in (
                        (option, element.find(option)) for option in options if element.find(option) is not None
                    ):
                        xml_dict[tag] = {option: value.text}
        return xml_dict


class Parser:
    """Class that represents a parser. This class translates the config created from a Template into a generic
    configuration dictionary

    Args:
        target: Target to parse the config from.
        config_globs: Glob patterns to obtain config files from.
        template: Template object to create a config from.
    """

    def __init__(self, target: Target, config_globs: list[str], template: Template):
        self.target = target
        self.template = template

        self.sections = self.template.sections
        self.options = self.template.options
        self.config_globs = config_globs

    def parse(self) -> defaultdict:
        """Returns a translated dictionary of network configuration properties.

        Returns:
            Dictionary containing network configuration properties for interface, dhcp, ips, gateways, dns, and netmask.
        """

        template = defaultdict(set)

        for path in self.expand_config_file_paths():
            config = self.template.create_config(path)

            for key, value in self.translate_network_config(config):
                if isinstance(value, list):
                    template[key].update(value)
                else:
                    template[key].add(value)
        return template

    def translate_network_config(self, config_dict: dict) -> list[tuple[str, Any]]:
        """Translates a parsed network configuration property to its generalized form:

        Returns:
            List containing the translated property and its value.
        """
        translated_values = []
        for section in self.sections:
            for option in self.options:
                value = self._get_option(config_dict, option, section)
                key = self.translate(value, option)
                if key:
                    translated_values.append((key, value))
        return translated_values

    def expand_config_file_paths(self) -> list[TargetPath]:
        """Expands all globbed config file path to discovery network configuration files.

        Returns:
            List containing the found network configuration files.
        """
        config_files = []

        for glob in self.config_globs:
            glob = glob.lstrip("/")
            config_files.extend(path for path in self.target.fs.path().glob(glob) if path.is_file())
        return config_files

    def translate(self, value: Any, option: str) -> str:
        """Translates the passed option value name to its corrensponding generalized option name.

        Returns:
            Translated option name.
        """
        translation_table = {
            "interface": ["name", "iface", "device"],
            "dhcp": ["bootproto", "dhcp", "dhcp4", "dhcpserver", "method"],
            "ips": ["ip", "address1", "addresses", "ipaddr", "address", "./address/local"],
            "netmask": ["netmask"],
            "gateway": ["gateway4", "gateway"],
            "dns": ["dns", "dns1", "dns-nameservers"],
        }

        for translation_key, translation_values in translation_table.items():
            if option in translation_values and value:
                return translation_key
        return None

    def _get_option(self, config: dict, option: str, section: str | None = None) -> str | Callable | None:
        """Internal function to get arbitrary options values from a parsed (non-translated) dictionary.

        Args:
            config: Configuration dictionary to obtain a option from.
            option: Option value to search for in the configuration dictionary.
            section: Section within the configuration dictionaty to look for the option value.

        Returns:
            Value(s) corrensponding to that network configuration option.
        """
        if not config:
            log.debug("Cannot get option %s: No config to parse", option)
            return None

        if section:
            # account for values of sections which are None
            config = config.get(section, {}) or {}

        for key, value in config.items():
            if key == option:
                return value
            if isinstance(value, dict):
                if option in value:
                    return value[option]
                return self._get_option(value, option)
        return None


class NetworkManager:
    """This class represents a Linux network managers on a given Linux based target. Detects if the network manager is
    active based on available configuration files and paths.

    Args:
        name: Name of the network manager to detect
        detection_globs: Glob patterns to detect network manager with.
        config_flobs: Glob patterns to retreive possible configuration files belonging to this network manager.
    """

    def __init__(self, name: str, detection_globs: tuple[str], config_globs: tuple[str]):
        self.target = None
        self.parser = None
        self.config = None

        self.name = name
        self.config_globs = (config_globs,) if isinstance(config_globs, str) else config_globs
        self.detection_globs = (detection_globs,) if isinstance(detection_globs, str) else detection_globs

    def detect(self, target: Target | None = None) -> bool:
        """Detects if the network manager is active on the target

        Returns:
            Whether a certain network manager is detected on the target
        """
        for path in self.detection_globs:
            path = path.lstrip("/")

            if next(target.fs.path().glob(path), None):
                target.log.debug("Found compatible network manager: %s", self.name)
                return True

        return False

    def register(self, target: Target, template: Template) -> bool:
        """Sets the detected parsing template and target the network manager.

        Args:
            target: Target object to register to this NetworkManager class.
            template: Parsing Template object to register to this NetworkManager class.

        Returns:
            Whether the registration process was executed succesfully.
        """
        self.target = target
        self.parser = Parser(target, self.config_globs, template)

        # Check if the unglobbed config files actually exists.
        # If not, nothing to parse. Thus we can skip this network manager.
        if self.parser.expand_config_file_paths():
            target.log.debug("Applying parsing template %s to network manager %s", template.name, self.name)
            target.log.debug("Registered network manager: %s as active", self.name)
            return True
        self.target, self.parser = None, None
        target.log.debug("Failed to register network manager %s as active", self.name)
        return False

    def parse(self) -> None:
        """Parse the network configuration for this network manager."""
        if self.registered:
            self.config = self.parser.parse()
        else:
            log.error("Network manager %s is not registered, cannot parse config", self.name)

    @property
    def interface(self) -> set:
        return self.config.get("interface")

    @property
    def ips(self) -> set:
        return NetworkManager.clean_ips(self.config.get("ips", set()))

    @property
    def dns(self) -> set:
        return NetworkManager.clean_ips(self.config.get("dns", set()))

    @property
    def dhcp(self) -> bool:
        # we do some post-processing on dhcp values to get a final boolean verdict
        return self._dhcp()

    @property
    def gateway(self) -> set:
        return NetworkManager.clean_ips(self.config.get("gateway", set()))

    @property
    def netmask(self) -> set:
        return NetworkManager.clean_ips(self.config.get("netmask", set()))

    @property
    def registered(self) -> bool:
        return bool(self.target and self.parser)

    def _dhcp(self) -> set:
        """Internal function to translate DHCP values to their boolean equivalent.

        Returns:
            Set of boolean values which indicate if DHCP is enabled for an interface.
        """
        # dhcp is either on or off per interface, so we require some extra post processing to give a accurate verdict.
        translation_table = {
            True: ("yes", "dhcp", "auto"),
            False: ("no", "none", "static", "manual"),
        }

        translated_value = set()

        if self.config.get("dhcp"):
            for dhcp_value in self.config.get("dhcp", ""):
                if isinstance(dhcp_value, bool):
                    return translated_value.add(dhcp_value)

                for key, value in translation_table.items():
                    if dhcp_value.lower() in value:
                        translated_value.add(key)
        else:
            translated_value.add(False)

        return translated_value

    @staticmethod
    def clean_ips(ips: Iterable) -> set:
        """Clean ip values before returning them.

        Args:
            ips: Iterable of ip addresses to clean.

        Returns:
            Set of cleaned ip addresses.
        """
        cleaned_ips = set()
        for ip_value in ips:
            # Remove broadcast and localhost ip addresses
            if should_ignore_ip(ip_value) or ip_value == "::":
                continue

            # Remove netmask cidr notation, eg. 1.2.3.4/24
            if "/" in ip_value:
                ip_value = ip_value.split("/")[0]

            # Remove encapsulated "'s
            if '"' in ip_value:
                ip_value = ip_value.replace('"', "")

            # Strip values
            ip_value = ip_value.strip()

            cleaned_ips.add(ip_value)

        return cleaned_ips

    def __repr__(self) -> str:
        return f"<NetworkManager {self.name}>"


class LinuxNetworkManager:
    """This class represents a collection of available network managers on a linux target as a collection of
    NetworkManager objects.

    Args:
        target: Target to discover and obtain network information from.
    """

    def __init__(self, target: Target):
        self.managers = []
        self.target = target

    def discover(self) -> None:
        """Discover which defined network managers are active on the target.

        Registers the discovered network managers as active for parsing later on.
        """
        for manager in MANAGERS:
            if manager.detect(self.target) and manager.register(self.target, TEMPLATES[manager.name]):
                self.managers.append(manager)

    def get_config_value(self, attr: str) -> list[set]:
        """Return the specified value from a network configuration option.

        Returns:
            List containing the values corrensponding to that configuration option.
        """
        values = []
        for manager in self.managers:
            manager.parse()
            value = getattr(manager, attr, None)
            if value:
                values.append(value)
        return values


def parse_unix_dhcp_log_messages(target: Target, iter_all: bool = False) -> set[str]:
    """Parse local syslog, journal and cloud init-log files for DHCP lease IPs.

    Args:
        target: Target to discover and obtain network information from.
        iter_all: Parse limited amount of journal messages (first 10000) or all of them.

    Returns:
        A set of found DHCP IP addresses.
    """
    ips = set()
    messages = set()

    for log_func in ["messages", "journal"]:
        try:
            messages = chain(messages, getattr(target, log_func)())
        except PluginError:  # noqa: PERF203
            target.log.debug("Could not search for DHCP leases in %s files", log_func)

    if not messages:
        target.log.warning("Could not search for DHCP leases using %s: No log entries found", log_func)

    for count, record in enumerate(messages):
        line = record.message

        if not line:
            continue

        # Ubuntu cloud-init
        if "Received dhcp lease on" in line:
            _, ip, _ = re.search(r"Received dhcp lease on (\w{0,}) for (\S+)\/(\S+)", line).groups()
            ips.add(ip)
            continue

        # Ubuntu DHCP
        if ("DHCPv4" in line or "DHCPv6" in line) and " address " in line and " via " in line:
            ip = line.split(" address ")[1].split(" via ")[0].strip().split("/")[0]
            ips.add(ip)
            continue

        # Ubuntu DHCP NetworkManager
        if "option ip_address" in line and ("dhcp4" in line or "dhcp6" in line) and "=> '" in line:
            ip = line.split("=> '")[1].replace("'", "").strip()
            ips.add(ip)
            continue

        # Debian and CentOS dhclient
        if hasattr(record, "service") and record.service == "dhclient" and "bound to" in line:
            ip = line.split("bound to")[1].split(" ")[1].strip()
            ips.add(ip)
            continue

        # CentOS DHCP and general NetworkManager
        if " address " in line and ("dhcp4" in line or "dhcp6" in line):
            ip = line.split(" address ")[1].strip()
            ips.add(ip)
            continue

        # Ubuntu/Debian DHCP networkd (Journal)
        if (
            hasattr(record, "code_func")
            and record.code_func == "dhcp_lease_acquired"
            and " address " in line
            and " via " in line
        ):
            _, ip, _, _ = re.search(r"^(\S+): DHCPv[4|6] address (\S+)\/(\S+) via (\S+)", line).groups()
            ips.add(ip)
            continue

        # The journal and syslog parsers can be relatively slow, so we stop when we have read 10000 entries,
        # or if we have found at least one ip address. When `iter_all` is `True` we continue searching.
        if not iter_all and (ips or count > 10_000):
            if not ips:
                target.log.warning("No DHCP IP addresses found in first 10000 journal or syslog entries")
            break

    return ips


def parse_unix_dhcp_leases(target: Target) -> set[str]:
    """Parse NetworkManager and dhclient DHCP ``.lease`` files.

    References:
        - https://linux.die.net/man/5/dhclient.conf

    Args:
        target: Target to discover and obtain network information from.

    Returns:
        A set of found DHCP IP addresses.
    """
    ips = set()

    for lease_file in chain(
        target.fs.path("/var/lib/NetworkManager").glob("*.lease*"),
        target.fs.path("/var/lib/dhcp").glob("*.lease*"),
        target.fs.path("/var/lib/dhclient").glob("*.lease*"),
    ):
        lease_text = lease_file.read_text()

        if "lease {" in lease_text:
            for line in lease_text.split("\n"):
                if "fixed-address" in line:
                    ips.add(line.split(" ")[-1].strip(";"))

        elif "ADDRESS=" in lease_text:
            lease = configutil.parse(lease_file, hint="env")
            if ip := lease.get("ADDRESS"):
                ips.add(ip)

    return ips


def should_ignore_ip(ip: str) -> bool:
    return any(ip.startswith(i) for i in IGNORED_IPS)


MANAGERS = [
    # Arch Linux netctl
    NetworkManager(
        "netctl",
        ("/etc/netctl/examples/*", "/usr/lib/systemd/system/netctl*.service"),
        ("/etc/netctl/*"),
    ),
    # NetworkManager is a python configuration tool for networking, for various Linux distro's
    NetworkManager(
        "NetworkManager",
        ("/usr/bin/nmcli", "/usr/bin/nmtui", "/etc/NetworkManager/NetworkManager.conf", "/usr/lib/NetworkManager/*"),
        ("/etc/NetworkManager/system-connections/*"),
    ),
    # Photon/Systemd Debian/Ubuntu network manager
    NetworkManager(
        "systemd-networkd",
        ("/etc/systemd/network/*.network", "/lib/systemd/system/systemd-networkd.servic*"),
        (
            "/etc/systemd/network/*.network",
            "/run/systemd/network/*.network",
            "/usr/lib/systemd/network/*.network",
        ),
    ),
    # Debian/Ubuntu WICD/Wicked network manager
    NetworkManager(
        "wicd",
        ("/etc/wicd/wire*-*.conf", "/usr/sbin/wicd", "/etc/dbus-1/system.d/wicd.*", "/etc/wicd/*.conf"),
        ("/etc/wicd/wire*-*.conf"),
    ),
    NetworkManager(
        "wicked",
        (
            "/etc/wicked/ifconfig/*.xml",
            "/usr/sbin/wicked*",
            "/usr/lib/systemd/system/wicked.*",
            "/etc/dbus-1/system.d/org.opensuse.Network.*",
        ),
        ("/etc/wicked/ifconfig/*.xml"),
    ),
    # Ubuntu netplan
    NetworkManager(
        "netplan",
        ("/etc/netplan/*.yaml", "/usr/sbin/netplan", "/usr/share/dbus-1/system.d/io.netplan.Netplan.conf"),
        ("/etc/netplan/*.yaml"),
    ),
    # CentOS/Red Hat/Fedora/SuSe sysconfig folder/files
    NetworkManager(
        "ifupdown",
        ("/etc/sysconfig/network-scripts/ifcfg-*", "/usr/sbin/ifup", "/usr/sbin/ifdown"),
        ("/etc/sysconfig/network-scripts/ifcfg-*", "/etc/sysconfig/network/ifcfg-*"),
    ),
    # Interfaces folder/files
    NetworkManager(
        "interfaces",
        ("/etc/network/interfaces", "/usr/sbin/ifup", "/usr/sbin/ifdown", "/usr/sbin/ifquery"),
        ("/etc/network/interfaces", "/etc/network/interfaces.d/*"),
    ),
]

TEMPLATES = {
    "netctl": Template(
        "netctl",
        ConfigParser(delimiters=("=", " "), comment_prefixes="#", dict_type=dict),
        ["netctl"],
        ["address", "gateway", "dns", "ip"],
    ),
    "netplan": Template(
        "netplan", YAML(typ="safe").load if HAS_YAML else None, ["network"], ["addresses", "dhcp4", "gateway4"]
    ),
    "NetworkManager": Template(
        "NetworkManager",
        ConfigParser(delimiters=("="), comment_prefixes="#", dict_type=dict),
        ["ipv4"],
        ["address1", "dns"],
    ),
    "systemd-networkd": Template(
        "systemd-networkd",
        ConfigParser(delimiters=("="), comment_prefixes=("#", ";"), dict_type=dict),
        ["Network", "Match"],
        ["address", "dhcp", "dns", "name", "dhcpserver"],
    ),
    "wicked": Template("wicked", ElementTree, ["ipv4:static", "ipv6:static", "name"], ["./address/local", "name"]),
    "ifupdown": Template(
        "ifupdown",
        ConfigParser(delimiters=("=", " "), comment_prefixes=("#"), dict_type=dict, strict=False),
        ["ifupdown"],
        ["ipaddr", "bootproto", "dns", "gateway", "name", "device", "dns1"],
    ),
    "interfaces": Template(
        "interfaces", None, ["interfaces"], ["iface", "address", "gateway", "netmask", "dns-nameservers"]
    ),
}



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from unittest.mock import Mock, patch

import pytest

from dissect.target.filesystems.jffs import JffsFilesystem, JffsFilesystemEntry

if TYPE_CHECKING:
    from collections.abc import Iterator


@pytest.fixture
def jffs_fs() -> Iterator[JffsFilesystem]:
    with patch("dissect.jffs.jffs2.JFFS2"):
        jffs_fs = JffsFilesystem(Mock())
        yield jffs_fs


@pytest.fixture
def jffs_fs_file_entry(jffs_fs: JffsFilesystem) -> JffsFilesystemEntry:
    raw_inode = Mock(uid=1000, guid=999, isize=165002)
    inode = Mock(
        mode=0o100664,
        inum=4,
        nlink=1,
        inode=raw_inode,
        atime=datetime(2024, 10, 1, 12, 0, 0, tzinfo=timezone.utc),
        mtime=datetime(2024, 10, 2, 12, 0, 0, tzinfo=timezone.utc),
        ctime=datetime(2024, 10, 3, 12, 0, 0, tzinfo=timezone.utc),
        is_file=lambda: True,
        is_dir=lambda: False,
        is_symlink=lambda: False,
    )

    return JffsFilesystemEntry(jffs_fs, "/some_file", inode)


@pytest.fixture
def jffs_fs_directory_entry(jffs_fs: JffsFilesystem) -> JffsFilesystemEntry:
    raw_inode = Mock(uid=1000, guid=999, isize=0)
    inode = Mock(
        mode=0o40775,
        inum=2,
        nlink=2,
        inode=raw_inode,
        atime=datetime(2024, 10, 1, 12, 0, 0, tzinfo=timezone.utc),
        mtime=datetime(2024, 10, 2, 12, 0, 0, tzinfo=timezone.utc),
        ctime=datetime(2024, 10, 3, 12, 0, 0, tzinfo=timezone.utc),
        is_file=lambda: False,
        is_dir=lambda: True,
        is_symlink=lambda: False,
    )

    return JffsFilesystemEntry(jffs_fs, "/some_directory", inode)


@pytest.mark.parametrize(
    ("entry_fixture", "expected_blocks"), [("jffs_fs_file_entry", 323), ("jffs_fs_directory_entry", 0)]
)
def test_jffs2_stat(entry_fixture: str, expected_blocks: int, request: pytest.FixtureRequest) -> None:
    jffs_entry: JffsFilesystemEntry = request.getfixturevalue(entry_fixture)
    stat = jffs_entry.stat()

    entry = jffs_entry.entry
    assert stat.st_mode == entry.mode
    assert stat.st_ino == entry.inum
    assert stat.st_dev == id(jffs_entry.fs)
    assert stat.st_nlink == entry.nlink
    assert stat.st_uid == entry.inode.uid
    assert stat.st_gid == entry.inode.gid
    assert stat.st_size == entry.inode.isize
    assert stat.st_atime == entry.atime.timestamp()
    assert stat.st_mtime == entry.mtime.timestamp()
    assert stat.st_ctime == entry.ctime.timestamp()
    assert stat.st_blksize == 4096
    assert stat.st_blocks == expected_blocks



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from flow.record import GroupedRecord

from dissect.target.exceptions import InvalidTaskError
from dissect.target.plugins.os.windows.tasks.records import (
    BaseTriggerRecord,
    DailyTriggerRecord,
    ExecRecord,
    MonthlyDateTriggerRecord,
    MonthlyDowTriggerRecord,
    PaddingTriggerRecord,
    WeeklyTriggerRecord,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath

atjob_def = """
struct PRIORITY {
    uint32  undefined1: 5;          /* bit 31..27 */
    uint32  normal : 1;             /* bit 26 - NORMAL_PRIORITY_CLASS */
    uint32  idle : 1;               /* bit 25 - IDLE_PRIORITY_CLASS */
    uint32  high : 1;               /* bit 24 - HIGH_PRIORITY_CLASS */
    uint32  realtime : 1;           /* bit 23 - REALTIME_PRIORITY_CLASS */
    uint32  undefined2 : 23;        /* bit 22..0 */
};

struct FLAGS {
    uint32  interactive : 1;        /* bit 31 - can interact with user. */
    uint32  delete_when_done : 1;   /* bit 30 - delete task when done. */
    uint32  disabled : 1;           /* bit 29 - task is disabled. */
    uint32  undefined3 : 1;         /* bit 28 */
    uint32  only_idle : 1;          /* bit 27 - only start when idle. */
    uint32  stop_on_idle_end : 1;   /* bit 26 - stop when no longer idle. */
    uint32  disallow_battery : 1;   /* bit 25 - don't start when on batteries. */
    uint32  stop_battery : 1;       /* bit 24 - stop when going to batteries. */
    uint32  docked : 1;             /* bit 23 - should be 0, unused. */
    uint32  hidden : 1;             /* bit 22 - hidden task. */
    uint32  internet_connected : 1; /* bit 21 - should be 0, unused. */
    uint32  restart_on_idle: 1;     /* bit 20 - restart task when returning to idle state. */
    uint32  wake_to_run : 1;        /* bit 19 - can resume or wake the system to run. */
    uint32  logged_on_only : 1;     /* bit 18 - only runs when specified user is logged on. */
    uint32  undefined2 : 10;        /* bit 8..17 */
    uint32  task_app_name_set : 1;  /* bit 7 - has app name. */
    uint32  undefined1 : 7;         /* bit 0..6 */
};

struct HRESULT {
    uint32  severity : 1;           /* 0 = success, 1 = failure. */
    uint32  reserved_value : 4;     /* reserved value */
    uint32  facility_code : 11;     /* responsibility for the error or warning. */
    uint32  return_code : 16;       /* error code that describes the error or warning. */
};

struct TFLAGS {
    uint32  has_end_date : 1;       /* bit 31 - stop at some point in time. */
    uint32  kill_at_end : 1;        /* bit 30 - stop at end of repetition period. */
    uint32  trigger_disabled : 1;   /* bit 29 - trigger is disabled. */
    uint32  unused : 29;            /* bit 28..0 - should be 0. */
};

struct TRIGGER {
    uint16  trigger_size;           /* trigger size, should be 0x0030. */
    uint16  reserved1;              /* reserved. */
    uint16  begin_year;             /* first trigger fire date, year. */
    uint16  begin_month;            /* first trigger fire date, month. */
    uint16  begin_day;              /* first trigger fire date, day. */
    uint16  end_year;               /* last trigger fire date, year. */
    uint16  end_month;              /* last trigger fire date, month. */
    uint16  end_day;                /* last trigger fire date, day. */
    uint16  start_hour;             /* hour of trigger fire. */
    uint16  start_minute;           /* minute of trigger fire. */
    uint32  minutes_duration;       /* task runs for duration in minutes. */
    uint32  minutes_interval;       /* task runs every interval in minutes. */
    TFLAGS  trigger_flags;          /* task trigger bit flags. */
    uint32  trigger_type;           /* trigger type. */
    uint16  trigger_specific0;      /* value specific to trigger type. */
    uint16  trigger_specific1;      /* value specific to trigger type. */
    uint16  trigger_specific2;      /* value specific to trigger type. */
    uint16  padding;                /* should be 0. */
    uint16  reserved2;              /* should be 0. */
    uint16  reserved3;              /* should be 0. */
};

struct ATJOB_DATA {
    uint16      windows_version;                    /* 0x00 - windows version that generated this task. */
    uint16      file_version;                       /* 0x02 - should be set to 1. */
    char        uuid[16];                           /* 0x04 - randomly generated UUID. */
    uint16      app_name_len_offset;                /* 0x14 - offset in bytes to app_name_len. */
    uint16      triggers_offset;                    /* 0x16 - offset in bytes to triggers. */
    uint16      retry_count;                        /* 0x18 - number of attempts to retry when failing. */
    uint16      retry_interval;                     /* 0x1a - minutes between retries. */
    uint16      idle_deadline;                      /* 0x1c - minutes to wait for idle machine. */
    uint16      idle_wait;                          /* 0x1e - minutes of idle before run task. */
    PRIORITY    task_prio;                          /* 0x20 - bit flags with max. one bit set. */
    uint32      max_run_time;                       /* 0x24 - milliseconds to wait for task complete. */
    uint32      exit_code;                          /* 0x28 - should be set to 0x00000000. */
    uint32      status;                             /* 0x2C - status value of the task. */
    FLAGS       task_flags;                         /* 0x30 - task flag bits. */
    uint16      last_year;                          /* 0x34 - last run year. */
    uint16      last_month;                         /* 0x36 - last run month. */
    uint16      last_weekday;                       /* 0x38 - last run weekday. */
    uint16      last_day;                           /* 0x3a - last run day of the month. */
    uint16      last_hour;                          /* 0x3c - last run hour (24h). */
    uint16      last_minute;                        /* 0x3e - last run minute. */
    uint16      last_second;                        /* 0x40 - last run second. */
    uint16      last_millisecond;                   /* 0x42 - last run millisecond. */
    uint16      running_instances;                  /* 0x44 - number of currently running instances. */
    uint16      app_name_len;                       /* 0x46 - app name character count. */
    wchar       app_name[app_name_len];             /* 0x48 - app name - null-terminated Unicode string. */
    uint16      par_char_count;                     /*      - parameters character count. */
    wchar       parameters[par_char_count];         /*      - parameters - null-terminated Unicode string. */
    uint16      dir_char_count;                     /*      - working dir character count. */
    wchar       working_dir[dir_char_count];        /*      - working dir - null-terminated Unicode string. */
    uint16      author_char_count;                  /*      - author character count. */
    wchar       author[author_char_count];          /*      - author - null-terminated Unicode string. */
    uint16      comment_char_count;                 /*      - comment character count. */
    wchar       comment[comment_char_count];        /*      - comment - null-terminated Unicode string. */
    uint16      user_data_size;                     /*      - user data size in bytes. */
    uint8       user_data[user_data_size];          /*      - arbitrary bits, implementation specific. */
    uint16      reserved_data_size;                 /*      - should be 0 or 8. */
    HRESULT     reserved_hresult;                   /*      - used to describe an error. */
    uint32      reserved_task_flags;                /*      - not used, should be zero. */
    uint16      trigger_count;                      /*      - size in bytes of array of triggers. */
    TRIGGER     task_triggers[trigger_count];       /*      - an arry of zero or more triggers. */
//      The following are optional fields and currently not parsed:
//      uint16      s_ver;                              /*      - SignatureVersion, should be 1. */
//      uint16      c_ver;                              /*      - MinClientVersion, should be 1. */
//      uint8       job_signature[64 * s_ver * c_ver];  /*      - calculated job signature. */
};
"""
c_atjob = cstruct().load(atjob_def)


class AtTask:
    """Initialize the class for opening .job task files created by at.exe.

    Args:
        job_file: the file to be parsed.
        target: the target system.
    """

    def __init__(self, job_file: TargetPath, tzinfo: datetime.tzinfo = datetime.timezone.utc):
        try:
            self.at_data = c_atjob.ATJOB_DATA(job_file.open())
        except Exception as e:
            raise InvalidTaskError(e)

        self.task_path = job_file
        self.tzinfo = tzinfo

        last_year = self.at_data.last_year
        last_month = self.at_data.last_month
        last_day = self.at_data.last_day
        last_hour = self.at_data.last_hour
        last_minute = self.at_data.last_minute
        last_second = self.at_data.last_second
        last_millisecond = self.at_data.last_millisecond

        # Create a datetime object using the variables
        if last_year == 0 or last_month == 0 or last_day == 0:
            self.date = None
        else:
            timestamp = datetime.datetime(
                year=last_year,
                month=last_month,
                day=last_day,
                hour=last_hour,
                minute=last_minute,
                second=last_second,
                microsecond=last_millisecond * 1000,  # Convert millisecond to microsecond
                tzinfo=self.tzinfo,
            )
            # Convert the datetime object to ISO format timestamp string
            self.last_run_date = timestamp.isoformat()

        self.author = self.at_data.author.rstrip("\x00")
        self.version = self.at_data.file_version
        self.description = self.at_data.comment.rstrip("\x00")
        self.restart_on_failure_interval = self.minutes_duration_to_iso(self.at_data.retry_interval)
        self.restart_on_failure_count = self.at_data.retry_count
        self.disallow_start_on_batteries = bool(self.at_data.task_flags.disallow_battery)
        self.stop_going_on_batteries = bool(self.at_data.task_flags.stop_battery)
        self.run_only_network_available = bool(self.at_data.task_flags.internet_connected)
        self.wake_to_run = bool(self.at_data.task_flags.wake_to_run)
        self.enabled = not self.at_data.task_flags.disabled
        self.hidden = bool(self.at_data.task_flags.hidden)
        self.idle_duration = self.minutes_duration_to_iso(self.at_data.idle_wait)
        self.idle_wait_timeout = self.minutes_duration_to_iso(self.at_data.idle_deadline)
        self.idle_stop_on_idle_end = bool(self.at_data.task_flags.stop_on_idle_end)
        self.idle_restart_on_idle = bool(self.at_data.task_flags.restart_on_idle)
        self.execution_time_limit = round(self.at_data.max_run_time / 60000)
        self.execution_time_limit = self.minutes_duration_to_iso(self.execution_time_limit)
        self.run_only_idle = bool(self.at_data.task_flags.only_idle)
        self.data = self.at_data.user_data

        # check which prio bit is set
        for key in self.at_data.task_prio.fields:
            if getattr(self.at_data.task_prio, key):
                self.priority = key

    def get_actions(self) -> Iterator[GroupedRecord]:
        """Get the at job task actions.

        Yields:
            An iterator of at job task actions.
        """
        action_type = "Exec"
        command = self.at_data.app_name.rstrip("\x00")
        args = self.at_data.parameters.rstrip("\x00")
        wrkdir = self.at_data.working_dir.rstrip("\x00")

        yield ExecRecord(
            action_type=action_type,
            command=command,
            arguments=args,
            working_directory=wrkdir,
        )

    def get_triggers(self) -> Iterator[GroupedRecord]:
        """Get the job task triggers.

        Yields:
            An iterator of at job task triggers.
        """
        TRIGGER_TYPE_NAMES = {
            0: "ONCE",
            1: "DAILY",
            2: "WEEKLY",
            3: "MONTHLYDATE",
            4: "MONTHLYDOW",
            5: "EVENT_ON_IDLE",
            6: "EVENT_AT_SYSTEMSTART",
            7: "EVENT_AT_LOGON",
        }
        for trigger in self.at_data.task_triggers:
            trigger_type = TRIGGER_TYPE_NAMES.get(trigger.trigger_type, None)
            trigger_enabled = not trigger.trigger_flags.trigger_disabled

            s_year = trigger.begin_year
            s_month = trigger.begin_month
            s_day = trigger.begin_day
            start_boundary = datetime.datetime(year=s_year, month=s_month, day=s_day, tzinfo=self.tzinfo)

            e_year = trigger.end_year
            e_month = trigger.end_month
            e_day = trigger.end_day
            end_boundary = None
            if trigger.trigger_flags.has_end_date:
                end_boundary = datetime.datetime(year=e_year, month=e_month, day=e_day, tzinfo=self.tzinfo)

            repetition_interval = self.minutes_duration_to_iso(trigger.minutes_interval)
            repetition_duration = self.minutes_duration_to_iso(trigger.minutes_duration)
            repetition_stop_duration_end = trigger.trigger_flags.kill_at_end == 1
            execution_time_limit = self.minutes_duration_to_iso(round(self.at_data.max_run_time / 60000))

            base = BaseTriggerRecord(
                trigger_enabled=trigger_enabled,
                start_boundary=start_boundary,
                end_boundary=end_boundary,
                repetition_interval=repetition_interval,
                repetition_duration=repetition_duration,
                repetition_stop_duration_end=repetition_stop_duration_end,
                execution_time_limit=execution_time_limit,
            )
            padding_record = PaddingTriggerRecord(
                padding=trigger.padding,
                reserved2=trigger.reserved2,
                reserved3=trigger.reserved3,
            )

            if trigger_type == "EVENT_AT_LOGON":
                # No trigger specific flags in job files for this trigger type
                pass

            if trigger_type == "EVENT_AT_SYSTEMSTART":
                # No trigger specific flags in job files for this trigger type
                pass

            if trigger_type == "EVENT_ON_IDLE":
                # No trigger specific flags in job files for this trigger type
                pass

            if trigger_type == "ONCE":
                # No trigger specific flags in job files for this trigger type
                pass

            if trigger_type == "DAILY":
                interval = trigger.trigger_specific0
                unused = [trigger.trigger_specific1, trigger.trigger_specific2]

                record = DailyTriggerRecord(
                    days_between_triggers=interval,
                    unused=unused,
                )

                yield GroupedRecord("filesystem/windows/task/daily", [base, record, padding_record])

            if trigger_type == "WEEKLY":
                interval = trigger.trigger_specific0

                # Find weekdays corresponding with trigger specific flags
                days_of_week = self.get_days_of_week(trigger.trigger_specific1)
                unused = [trigger.trigger_specific2]

                record = WeeklyTriggerRecord(
                    weeks_between_triggers=interval,
                    days_of_week=days_of_week,
                    unused=unused,
                )

                yield GroupedRecord("filesystem/windows/task/weekly", [base, record, padding_record])

            if trigger_type == "MONTHLYDATE":
                # Convert trigger_specific fields to binary and pad with zeroes to 16 digits
                day_flag_part1 = f"{trigger.trigger_specific0:b}".zfill(16)
                day_flag_part2 = f"{trigger.trigger_specific1:b}".zfill(16)
                # Concatenate the two binary strings to form the complete day_flag
                day_of_month_flag = day_flag_part2 + day_flag_part1
                # Reverse the day_flag string (LE), find the index of "1", add 1 to get the day of the month
                day_of_month = day_of_month_flag[::-1].index("1") + 1

                # Find months corresponding with trigger specific flags
                months_of_year = self.get_months_of_year(trigger.trigger_specific2)

                record = MonthlyDateTriggerRecord(
                    day_of_month=[day_of_month],
                    months_of_year=months_of_year,
                )

                yield GroupedRecord("filesystem/windows/task/monthly_date", [base, record, padding_record])

            if trigger_type == "MONTHLYDOW":
                week = trigger.trigger_specific0
                days = self.get_days_of_week(trigger.trigger_specific1)
                months = self.get_months_of_year(trigger.trigger_specific2)
                record = MonthlyDowTriggerRecord(
                    which_week=[week],
                    days_of_week=days,
                    months_of_year=months,
                )

                yield GroupedRecord("filesystem/windows/task/monthly_dow", [base, record, padding_record])

    def minutes_duration_to_iso(self, minutes: int) -> str | None:
        """Convert the given number of minutes to an ISO 8601 duration format string, like those found in the xml tasks.
        The most significant unit is days (D), the least significant is minutes (M).

        Args:
            minutes: The number of minutes to convert.

        Returns:
            An ISO 8601 duration format string representing the given number of minutes,
            or ``None`` if the number of minutes is zero.

        Raises:
            TypeError: If the minutes argument is not an integer.
        """
        if not isinstance(minutes, int):
            raise TypeError("Expected an integer for the minutes argument")

        if minutes == 0:
            return None
        # Calculate the number of days, hours, and minutes
        days, minutes = divmod(minutes, 1440)
        hours, minutes = divmod(minutes, 60)

        # Build the ISO duration format string
        duration_iso = "P" if days == 0 else f"P{days}D"
        duration_iso += "T" if hours or minutes else ""
        duration_iso += f"{hours}H" if hours != 0 else ""
        duration_iso += f"{minutes}M" if minutes != 0 else ""

        return duration_iso

    def get_flags_data(self, flags: int, items: list[str]) -> Iterator[str]:
        """Create a generator of items corresponding to the flags.

        Args:
            flags: An integer representing the trigger specific flags.
            items: A list of items corresponding to the flags.

        Yields:
            Each item based on the flags.
        """
        flags_binary = f"{flags:b}"  # Convert to binary
        for i, char in enumerate(flags_binary[::-1]):
            if char == "1":
                yield items[i]

    def get_months_of_year(self, flags: int) -> list[str]:
        """Convert 16-bit flags to a list of months of the year.

        Args:
            flags: An integer representing the trigger specific flags. See also:
            https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-tsch/5ba70e9b-c0f1-49f6-9aae-b52231346108

        Returns:
            A list of months as strings.
        """
        months = [
            "January",
            "February",
            "March",
            "April",
            "May",
            "June",
            "July",
            "August",
            "September",
            "October",
            "November",
            "December",
        ]
        return list(self.get_flags_data(flags, months))

    def get_days_of_week(self, flags: int) -> list[str]:
        """Get the list of weekdays corresponding to the given trigger specific 16-bit flags.

        Args:
            flags: An integer representing the trigger specific flags. See also:
            https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-tsch/b7a0fc39-b43a-435a-9f37-60e48f340b9b

        Returns:
            A list of weekdays as strings.
        """
        weekdays = [
            "Sunday",
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
        ]
        return list(self.get_flags_data(flags, weekdays))



from __future__ import annotations

from io import BytesIO

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.plugins.child.esxi import ESXiChildTargetPlugin
from dissect.target.plugins.os.unix.esxi._os import ESXiPlugin
from dissect.target.target import Target


def test_esxi_children() -> None:
    """Test that the ESXi child target plugin lists children correctly."""
    vfs = VirtualFilesystem()
    vfs.map_file_fh(
        "/etc/vmware/hostd/vmInventory.xml",
        BytesIO(b"""
          <ConfigRoot>
            <ConfigEntry id="0000">
              <objID>1</objID>
              <secDomain/>
              <vmxCfgPath>/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Alpine/Alpine.vmx</vmxCfgPath>
            </ConfigEntry>
          </ConfigRoot>
        """),
    )
    vfs.map_file_fh(
        "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Alpine/Alpine.vmx", BytesIO(b'displayName = "Alpine"')
    )

    # vmx not defined in vmInventory
    vfs.map_file_fh(
        "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Debian/Debian.vmx", BytesIO(b'displayName = "Debian"')
    )

    target = Target()
    target._os_plugin = ESXiPlugin(target)
    target.filesystems.add(vfs)
    target.fs.mount("/", vfs)
    target.apply()

    target.add_plugin(ESXiChildTargetPlugin)

    children = [child for _, child in target.list_children()]

    assert len(children) == 2

    assert children[0].type == "esxi"
    assert children[0].name == "Alpine"
    assert str(children[0].path) == "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Alpine/Alpine.vmx"

    assert children[1].type == "esxi"
    assert children[1].name == "Debian"
    assert str(children[1].path) == "/vmfs/volumes/6800e48c-3dcffc58-6af7-bc2411ec8065/Debian/Debian.vmx"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.nscollector import NsCollectorTarSubLoader
from dissect.target.loaders.tar import TarLoader
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``NsCollectorTarSubLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/nscollector/collector_P_10.164.0.3_22Oct2025_11_31.tar.gz")

    target = opener(path)
    assert isinstance(target._loader, TarLoader)
    assert isinstance(target._loader.subloader, NsCollectorTarSubLoader)
    assert target.path == path


def test_compressed_tar(caplog: pytest.LogCaptureFixture) -> None:
    """Test if we map a compressed NetScaler Collector tar image correctly."""
    path = absolute_path("_data/loaders/nscollector/collector_P_10.164.0.3_22Oct2025_11_31.tar.gz")

    with caplog.at_level(logging.WARNING):
        loader = loader_open(path)
        assert isinstance(loader, TarLoader)
        assert "is compressed" in caplog.text

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, NsCollectorTarSubLoader)
    assert len(t.filesystems) == 1

    t.apply()
    test_file = t.fs.path("/nsconfig/ns.conf")
    assert test_file.exists()
    assert test_file.is_file()
    assert test_file.open().readline() == b"#NS14.1 Build 51.80\n"
